Package: PoissonBinomial
Type: Package
Title: Efficient Computation of Ordinary and Generalised Poisson
        Binomial Distributions
Version: 1.2.8
Date: 2026-03-06
Authors@R: person(
    "Florian", "Junge", 
    role = c("aut", "cre"), 
    email = "florian.junge@mailbox.org",
    comment = c(ORCID = "0009-0001-6856-6938")
  )
Maintainer: Florian Junge <florian.junge@mailbox.org>
Language: en-GB
Description: Efficient implementations of multiple exact and approximate methods
  as described in Hong (2013) <doi:10.1016/j.csda.2012.10.006>, Biscarri, Zhao &
  Brunner (2018) <doi:10.1016/j.csda.2018.01.007> and Zhang, Hong & Balakrishnan
  (2018) <doi:10.1080/00949655.2018.1440294> for computing the probability mass,
  cumulative distribution and quantile functions, as well as generating random
  numbers for both the ordinary and generalised Poisson binomial distribution.
License: GPL-3
Encoding: UTF-8
Imports: Rcpp (>= 1.0.11)
LinkingTo: Rcpp
SystemRequirements: fftw3 (>= 3.3)
Suggests: knitr, rmarkdown, microbenchmark
VignetteBuilder: knitr
URL: https://github.com/fj86/PoissonBinomial
BugReports: https://github.com/fj86/PoissonBinomial/issues
RoxygenNote: 7.3.3
NeedsCompilation: yes
Packaged: 2026-03-06 13:14:07 UTC; fjunge
Author: Florian Junge [aut, cre] (ORCID:
    <https://orcid.org/0009-0001-6856-6938>)
Repository: CRAN
Date/Publication: 2026-03-06 13:50:13 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-04-26 18:07:11 UTC; unix
Archs: PoissonBinomial.so.dSYM

# Generated by roxygen2: do not edit by hand

export(dgpbinom)
export(dpbinom)
export(pgpbinom)
export(ppbinom)
export(qgpbinom)
export(qpbinom)
export(rgpbinom)
export(rpbinom)
import(Rcpp)
importFrom(stats,dbinom)
importFrom(stats,pbinom)
importFrom(stats,rbinom)
importFrom(stats,runif)
importFrom(stats,stepfun)
useDynLib(PoissonBinomial, .registration = TRUE)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo = FALSE------------------------------------------------------
library(PoissonBinomial)

## ----pa1----------------------------------------------------------------------
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Poisson")
ppbinom(NULL, pp, wt, "Poisson")

## ----pa2----------------------------------------------------------------------
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "Poisson")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Poisson") - dpbinom(NULL, pp))

# U(0, 0.01) random probabilities of success
pp <- runif(20, 0, 0.01)
dpbinom(NULL, pp, method = "Poisson")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Poisson") - dpbinom(NULL, pp))

## ----am1----------------------------------------------------------------------
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
mean(rep(pp, wt))

dpbinom(NULL, pp, wt, "Mean")
ppbinom(NULL, pp, wt, "Mean")

## ----am2----------------------------------------------------------------------
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "Mean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Mean") - dpbinom(NULL, pp))

# U(0.3, 0.5) random probabilities of success
pp <- runif(20, 0.3, 0.5)
dpbinom(NULL, pp, method = "Mean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Mean") - dpbinom(NULL, pp))

# U(0.39, 0.41) random probabilities of success
pp <- runif(20, 0.39, 0.41)
dpbinom(NULL, pp, method = "Mean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Mean") - dpbinom(NULL, pp))

## ----gma1---------------------------------------------------------------------
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
prod(rep(pp, wt))^(1/sum(wt))

dpbinom(NULL, pp, wt, "GeoMean")
ppbinom(NULL, pp, wt, "GeoMean")

## ----gma2---------------------------------------------------------------------
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "GeoMean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMean") - dpbinom(NULL, pp))

# U(0.4, 0.6) random probabilities of success
pp <- runif(20, 0.4, 0.6)
dpbinom(NULL, pp, method = "GeoMean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMean") - dpbinom(NULL, pp))

# U(0.49, 0.51) random probabilities of success
pp <- runif(20, 0.49, 0.51)
dpbinom(NULL, pp, method = "GeoMean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMean") - dpbinom(NULL, pp))

## ----gmb1---------------------------------------------------------------------
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
1 - prod(1 - rep(pp, wt))^(1/sum(wt))

dpbinom(NULL, pp, wt, "GeoMeanCounter")
ppbinom(NULL, pp, wt, "GeoMeanCounter")

## ----gmb2---------------------------------------------------------------------
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "GeoMeanCounter")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMeanCounter") - dpbinom(NULL, pp))

# U(0.4, 0.6) random probabilities of success
pp <- runif(20, 0.4, 0.6)
dpbinom(NULL, pp, method = "GeoMeanCounter")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMeanCounter") - dpbinom(NULL, pp))

# U(0.49, 0.51) random probabilities of success
pp <- runif(20, 0.49, 0.51)
dpbinom(NULL, pp, method = "GeoMeanCounter")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMeanCounter") - dpbinom(NULL, pp))

## ----na1-ord------------------------------------------------------------------
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Normal")
ppbinom(NULL, pp, wt, "Normal")

## ----na2-ord------------------------------------------------------------------
set.seed(1)

# 10 random probabilities of success
pp <- runif(10)
dpn <- dpbinom(NULL, pp, method = "Normal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
dpn <- dpbinom(NULL, pp, method = "Normal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100000 random probabilities of success
pp <- runif(100000)
dpn <- dpbinom(NULL, pp, method = "Normal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

## ----rna1-ord-----------------------------------------------------------------
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "RefinedNormal")
ppbinom(NULL, pp, wt, "RefinedNormal")

## ----rna2-ord-----------------------------------------------------------------
set.seed(1)

# 10 random probabilities of success
pp <- runif(10)
dpn <- dpbinom(NULL, pp, method = "RefinedNormal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
dpn <- dpbinom(NULL, pp, method = "RefinedNormal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100000 random probabilities of success
pp <- runif(100000)
dpn <- dpbinom(NULL, pp, method = "RefinedNormal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

## ----benchmark-ord------------------------------------------------------------
library(microbenchmark)
set.seed(1)

f1 <- function() dpbinom(NULL, runif(4000), method = "Normal")
f2 <- function() dpbinom(NULL, runif(4000), method = "Poisson")
f3 <- function() dpbinom(NULL, runif(4000), method = "RefinedNormal")
f4 <- function() dpbinom(NULL, runif(4000), method = "Mean")
f5 <- function() dpbinom(NULL, runif(4000), method = "GeoMean")
f6 <- function() dpbinom(NULL, runif(4000), method = "GeoMeanCounter")
f7 <- function() dpbinom(NULL, runif(4000), method = "DivideFFT")

microbenchmark(f1(), f2(), f3(), f4(), f5(), f6(), f7(), times = 51)

## ----na1-gen------------------------------------------------------------------
set.seed(2)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)

dgpbinom(NULL, pp, va, vb, wt, "Normal")
pgpbinom(NULL, pp, va, vb, wt, "Normal")

## ----na2-gen------------------------------------------------------------------
set.seed(2)

# 10 random probabilities of success
pp <- runif(10)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "Normal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100 random probabilities of success
pp <- runif(100)
va <- sample(0:100, 100, TRUE)
vb <- sample(0:100, 100, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "Normal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
va <- sample(0:1000, 1000, TRUE)
vb <- sample(0:1000, 1000, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "Normal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

## ----rna1-gen-----------------------------------------------------------------
set.seed(2)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)
dgpbinom(NULL, pp, va, vb, wt, "RefinedNormal")
pgpbinom(NULL, pp, va, vb, wt, "RefinedNormal")

## ----rna2-gen-----------------------------------------------------------------
set.seed(2)

# 10 random probabilities of success
pp <- runif(10)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100 random probabilities of success
pp <- runif(100)
va <- sample(0:100, 100, TRUE)
vb <- sample(0:100, 100, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
va <- sample(0:1000, 1000, TRUE)
vb <- sample(0:1000, 1000, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

## ----benchmark-gen------------------------------------------------------------
library(microbenchmark)
n <- 1500
set.seed(2)
va <- sample(1:50, n, TRUE)
vb <- sample(1:50, n, TRUE)

f1 <- function() dgpbinom(NULL, runif(n), va, vb, method = "Normal")
f2 <- function() dgpbinom(NULL, runif(n), va, vb, method = "RefinedNormal")
f3 <- function() dgpbinom(NULL, runif(n), va, vb, method = "DivideFFT")

microbenchmark(f1(), f2(), f3(), times = 51)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup, echo = FALSE------------------------------------------------------
library(PoissonBinomial)

## ----ex-opdb------------------------------------------------------------------
# Case 1
dpbinom(NULL, rep(0.3, 7))
dbinom(0:7, 7, 0.3) # equal results

dpbinom(NULL, c(0, 0, 0, 0, 0, 0, 0)) # only 0 is observable
dpbinom(0, c(0, 0, 0, 0, 0, 0, 0)) # confirmation

dpbinom(NULL, c(1, 1, 1, 1, 1, 1, 1)) # only 7 is observable
dpbinom(7, c(1, 1, 1, 1, 1, 1, 1)) # confirmation

# Case 2
dpbinom(NULL, c(0, 0, 0, 0, 1, 1, 1)) # only 3 is observable
dpbinom(3, c(0, 0, 0, 0, 1, 1, 1)) # confirmation

# Case 3
dpbinom(NULL, c(0, 0, 0.1, 0.2, 0.4, 0.8, 1)) # only 1-5 are observable
dpbinom(1:5, c(0, 0, 0.1, 0.2, 0.4, 0.8, 1)) # confirmation

dpbinom(NULL, c(0, 0, 0.4, 1)) # only 1 and 2 are observable
dpbinom(1:2, c(0, 0, 0.4, 1)) # confirmation

## ----ex-gpdb------------------------------------------------------------------
set.seed(1)
pp <- runif(7)
va <- sample(0:6, 7, TRUE)
vb <- sample(0:6, 7, TRUE)

# Case 1
dgpbinom(NULL, pp, rep(1, 7), rep(0, 7))
dpbinom(NULL, pp) # equal results

dgpbinom(NULL, pp, rep(0, 7), rep(1, 7))
dpbinom(NULL, 1 - pp) # equal results

dgpbinom(NULL, pp, c(rep(1, 3), rep(0, 4)), c(rep(0, 3), rep(1, 4)))
dpbinom(NULL, c(pp[1:3], 1 - pp[4:7])) # reorder for 0 and 1; equal results

# Case 2 a)
dgpbinom(NULL, rep(0, 7), rep(4, 7), rep(2, 7)) # only 14 is observable
dgpbinom(7 * 2, rep(0, 7), rep(4, 7), rep(2, 7)) # confirmation

# Case 2 b)
dgpbinom(NULL, rep(1, 7), rep(4, 7), rep(2, 7)) # only 28 is observable
dgpbinom(7 * 4, rep(1, 7), rep(4, 7), rep(2, 7)) # confirmation

# Case 2 c)
dgpbinom(NULL, rep(0.3, 7), rep(4, 7), rep(2, 7))
dbinom(0:7, 7, 0.3) # equal results, but on different support set

# Case 2 d)
dgpbinom(NULL, pp, rep(4, 7), rep(2, 7))
dpbinom(NULL, pp) # equal results, but on different support set

# Case 3 a)
dgpbinom(NULL, c(0, 0, 0, 0, 0, 0, 0), va, vb) # only sum(vb) is observable
dgpbinom(sum(vb), rep(0, 7), va, vb) # confirmation

# Case 3 b)
dgpbinom(NULL, c(1, 1, 1, 1, 1, 1, 1), va, vb) # only sum(va) is observable
dgpbinom(sum(va), rep(1, 7), va, vb) # confirmation

# Case 3 c)
dgpbinom(NULL, c(0, 0, 0, 1, 1, 1, 1), va, vb) # only sum(va[4:7], vb[1:3]) is observable
dgpbinom(sum(va[4:7], vb[1:3]), c(0, 0, 0, 1, 1, 1, 1), va, vb) # confirmation

# Case 4
dgpbinom(NULL, c(0, 0, 0.3, 0.6, 1, 1, 1), va, vb)
sure <- sum(va[5:7], vb[1:2])
x.transf <- sum(pmin(va[3:4], vb[3:4])):sum(pmax(va[3:4], vb[3:4]))
dgpbinom(sure + x.transf, c(0, 0, 0.3, 0.6, 1, 1, 1), va, vb)
dgpbinom(x.transf, c(0.3, 0.6), va[3:4], vb[3:4]) # equal results

dgpbinom(NULL, c(0, 0, 0, 0.6, 1, 1, 1), va, vb)
sure <- sum(va[5:7], vb[1:3])
x.transf <- va[4]:vb[4]
dgpbinom(sure + x.transf, c(0, 0, 0, 0.6, 1, 1, 1), va, vb)
dgpbinom(x.transf, 0.6, va[4], vb[4]) # equal results; essentially transformed Bernoulli

#' @name PoissonBinomial-Distribution
#'
#' @importFrom stats dbinom pbinom runif
#'
#' @title The Poisson Binomial Distribution
#'
#' @description
#' Density, distribution function, quantile function and random generation for
#' the Poisson binomial distribution with probability vector \code{probs}.
#'
#' @param x           Either a vector of observed numbers of successes or NULL.
#'                    If NULL, probabilities of all possible observations are
#'                    returned.
#' @param p           Vector of probabilities for computation of quantiles.
#' @param n           Number of observations. If \code{length(n) > 1}, the
#'                    length is taken to be the number required.
#' @param probs       Vector of probabilities of success of each Bernoulli
#'                    trial.
#' @param method      Character string that specifies the method of computation
#'                    and must be one of \code{"DivideFFT"}, \code{"Convolve"},
#'                    \code{"Characteristic"}, \code{"Recursive"},
#'                    \code{"Mean"}, \code{"GeoMean"}, \code{"GeoMeanCounter"},
#'                    \code{"Poisson"}, \code{"Normal"} or
#'                    \code{"RefinedNormal"} (abbreviations are allowed).
#' @param wts         Vector of non-negative integer weights for the input
#'                    probabilities.
#' @param log,log.p   Logical value indicating if results are given as
#'                    logarithms.
#' @param lower.tail  Logical value indicating if results are \eqn{P[X \leq x]}
#'                    (if \code{TRUE}; default) or \eqn{P[X > x]} (if 
#'                    \code{FALSE}).
#' @param generator   Character string that specifies the random number
#'                    generator and must either be \code{"Sample"} (default) or
#'                    \code{"Bernoulli"} (abbreviations are allowed). See
#'                    Details for more information.
#'
#' @details
#' See the references for computational details. The \emph{Divide and Conquer}
#' (\code{"DivideFFT"}) and \emph{Direct Convolution} (\code{"Convolve"})
#' algorithms are derived and described in Biscarri, Zhao & Brunner (2018). The
#' \emph{Discrete Fourier Transformation of the Characteristic Function}
#' (\code{"Characteristic"}), the \emph{Recursive Formula} (\code{"Recursive"}),
#' the \emph{Poisson Approximation} (\code{"Poisson"}), the
#' \emph{Normal Approach} (\code{"Normal"}) and the
#' \emph{Refined Normal Approach} (\code{"RefinedNormal"}) are described in Hong
#' (2013). The calculation of the \emph{Recursive Formula} was modified to
#' overcome the excessive memory requirements of Hong's implementation.
#'
#' The \code{"Mean"} method is a naive binomial approach using the arithmetic
#' mean of the probabilities of success. Similarly, the \code{"GeoMean"} and
#' \code{"GeoMeanCounter"} procedures are binomial approximations, too, but
#' they form the geometric mean of the probabilities of success
#' (\code{"GeoMean"}) and their counter probabilities (\code{"GeoMeanCounter"}),
#' respectively.
#'
#' In some special cases regarding the values of \code{probs}, the \code{method}
#' parameter is ignored (see Introduction vignette).
#'
#' Random numbers can be generated in two ways. The \code{"Sample"} method
#' uses \code{R}'s \code{sample} function to draw random values according to
#' their probabilities that are calculated by \code{dgpbinom}. The
#' \code{"Bernoulli"} procedure ignores the \code{method} parameter and
#' simulates Bernoulli-distributed random numbers according to the probabilities
#' in \code{probs} and sums them up. It is a bit slower than the \code{"Sample"}
#' generator, but may yield better results, as it allows to obtain observations
#' that cannot be generated by the \code{"Sample"} procedure, because
#' \code{dgpbinom} may compute 0-probabilities, due to rounding, if the length
#' of \code{probs} is large and/or its values contain a lot of very small
#' values.
#'
#' @return
#' \code{dpbinom} gives the density, \code{ppbinom} computes the distribution
#' function, \code{qpbinom} gives the quantile function and \code{rpbinom}
#' generates random deviates.
#'
#' For \code{rpbinom}, the length of the result is determined by \code{n}, and
#' is the lengths of the numerical arguments for the other functions.
#'
#' @references
#' Hong, Y. (2013). On computing the distribution function for the Poisson
#'   binomial distribution. \emph{Computational Statistics & Data Analysis},
#'   \strong{59}, pp. 41-51. \doi{10.1016/j.csda.2012.10.006}
#'
#' Biscarri, W., Zhao, S. D. and Brunner, R. J. (2018) A simple and fast method
#'   for computing the Poisson binomial distribution.
#'   \emph{Computational Statistics and Data Analysis}, \strong{31}, pp.
#'   216–222. \doi{10.1016/j.csda.2018.01.007}
#'
#' @examples
#' set.seed(1)
#' pp <- c(0, 0, runif(995), 1, 1, 1)
#' qq <- seq(0, 1, 0.01)
#'
#' dpbinom(NULL, pp, method = "DivideFFT")
#' ppbinom(450:550, pp, method = "DivideFFT")
#' qpbinom(qq, pp, method = "DivideFFT")
#' rpbinom(100, pp, method = "DivideFFT")
#'
#' dpbinom(NULL, pp, method = "Convolve")
#' ppbinom(450:550, pp, method = "Convolve")
#' qpbinom(qq, pp, method = "Convolve")
#' rpbinom(100, pp, method = "Convolve")
#'
#' dpbinom(NULL, pp, method = "Characteristic")
#' ppbinom(450:550, pp, method = "Characteristic")
#' qpbinom(qq, pp, method = "Characteristic")
#' rpbinom(100, pp, method = "Characteristic")
#'
#' dpbinom(NULL, pp, method = "Recursive")
#' ppbinom(450:550, pp, method = "Recursive")
#' qpbinom(qq, pp, method = "Recursive")
#' rpbinom(100, pp, method = "Recursive")
#'
#' dpbinom(NULL, pp, method = "Mean")
#' ppbinom(450:550, pp, method = "Mean")
#' qpbinom(qq, pp, method = "Mean")
#' rpbinom(100, pp, method = "Mean")
#'
#' dpbinom(NULL, pp, method = "GeoMean")
#' ppbinom(450:550, pp, method = "GeoMean")
#' qpbinom(qq, pp, method = "GeoMean")
#' rpbinom(100, pp, method = "GeoMean")
#'
#' dpbinom(NULL, pp, method = "GeoMeanCounter")
#' ppbinom(450:550, pp, method = "GeoMeanCounter")
#' qpbinom(qq, pp, method = "GeoMeanCounter")
#' rpbinom(100, pp, method = "GeoMeanCounter")
#'
#' dpbinom(NULL, pp, method = "Poisson")
#' ppbinom(450:550, pp, method = "Poisson")
#' qpbinom(qq, pp, method = "Poisson")
#' rpbinom(100, pp, method = "Poisson")
#'
#' dpbinom(NULL, pp, method = "Normal")
#' ppbinom(450:550, pp, method = "Normal")
#' qpbinom(qq, pp, method = "Normal")
#' rpbinom(100, pp, method = "Normal")
#'
#' dpbinom(NULL, pp, method = "RefinedNormal")
#' ppbinom(450:550, pp, method = "RefinedNormal")
#' qpbinom(qq, pp, method = "RefinedNormal")
#' rpbinom(100, pp, method = "RefinedNormal")
#'
#' @export
dpbinom <- function(x, probs, wts = NULL, method = "DivideFFT", log = FALSE) {
  ## preliminary checks
  # number of probabilities
  n <- length(probs)
  
  # check if 'x' contains only integers
  if(!is.null(x) && any(x - round(x) != 0)) {
    warning("'x' should contain integers only! Using rounded off values.")
    x <- floor(x)
  }
  
  # check if 'probs' contains only probabilities
  if(is.null(probs) || any(is.na(probs) | probs < 0 | probs > 1))
    stop("'probs' must contain real numbers between 0 and 1!")
  
  # make sure that the value of 'method' matches an implemented procedure
  method <- match.arg(
    arg = method, 
    choices = c(
      "DivideFFT", "Convolve", "Characteristic", "Recursive", "Mean",
      "GeoMean", "GeoMeanCounter", "Poisson", "Normal", "RefinedNormal"
    )
  )
  
  # check if 'wts' contains only integers (zeros are allowed)
  if(!is.null(wts) && any(is.na(wts) | wts < 0 | abs(wts - round(wts)) > 1e-07))
    stop("'wts' must contain non-negative integers!")
  
  if(!is.null(wts) && length(wts) != n)
    stop("'probs' and 'wts' (if not NULL) must have the same length!")
  
  ## expand 'probs' according to the counts in 'wts'
  # if 'wts' is NULL, set it to be a vector of ones
  if(is.null(wts)) wts <- rep(1, n)
  
  # expand 'probs'
  probs <- rep(probs, wts)
  
  # re-compute length of 'probs' (= sum of 'wts')
  n <- sum(wts)
  
  # if x = NULL, return all possible probabilities
  if(is.null(x)) x <- 0:n
  
  # identify valid 'x' values (invalid ones will have 0-probability)
  idx.x <- which(x >= 0 & x <= n)
  
  # select valid observations
  y <- x[idx.x]
  
  ## compute probabilities
  # vector for storing the probabilities
  d <- double(length(x))
  
  # no computation needed, if there are no valid observations in 'x'
  if(length(idx.x)) {
    # which probabilities are 0 or 1
    idx0  <- which(probs == 0)
    idx1  <- which(probs == 1)
    probs <- probs[probs > 0 & probs < 1]
    
    # number of zeros and ones
    n0 <- length(idx0)
    n1 <- length(idx1)
    np <- n - n0 - n1
    
    # relevant observations
    idx.y <- which(y %in% n1:(n - n0))
    
    if(length(idx.y)) {
      z <- y[idx.y] - n1
    
      if(np == 0) {
        # 'probs' contains only zeros and ones => only one possible observation
        d[idx.x][idx.y] <- 1
      } else {
        if(np == 1) {
          # 'probs' contains only one value in (0, 1) => Bernoulli distribution
          d[idx.x][idx.y] <- c(1 - probs, probs)[z + 1]
        } else {
          if(all(probs == probs[1])) {
            # all values of 'probs' are equal => standard binomial distribution
            d[idx.x][idx.y] <- dbinom(z, np, probs[1])
          } else {
            # otherwise, compute distribution according to 'method'
            d[idx.x][idx.y] <- switch(
              EXPR = method,
              DivideFFT = dpb_dc(z, probs),
              Convolve = dpb_conv(z, probs),
              Characteristic = dpb_dftcf(z, probs),
              Recursive = dpb_rf(z, probs),
              Mean = dpb_mean(z, probs),
              GeoMean = dpb_gmba(z, probs, FALSE),
              GeoMeanCounter = dpb_gmba(z, probs, TRUE),
              Poisson = dpb_pa(z, probs),
              Normal = dpb_na(z, probs, FALSE),
              RefinedNormal = dpb_na(z, probs, TRUE)
            )
          }
        }
      }
    }
  }
  
  # logarithm, if required
  if(log) d <- log(d)
  
  # return results
  return(d)
}

#' @rdname PoissonBinomial-Distribution
#' @export
ppbinom <- function(
  x,
  probs,
  wts = NULL,
  method = "DivideFFT",
  lower.tail = TRUE,
  log.p = FALSE
) {
  ## preliminary checks
  # number of probabilities
  n <- length(probs)
  
  # check if 'x' contains only integers
  if(!is.null(x) && any(x - round(x) != 0)) {
    warning("'x' should contain integers only! Using rounded off values.")
    x <- floor(x)
  }
  
  # check if 'probs' contains only probabilities
  if(is.null(probs) || any(is.na(probs) | probs < 0 | probs > 1))
    stop("'probs' must contain real numbers between 0 and 1!")
  
  # make sure that the value of 'method' matches an implemented procedure
  method <- match.arg(
    arg = method, 
    choices = c(
      "DivideFFT", "Convolve", "Characteristic", "Recursive", "Mean",
      "GeoMean", "GeoMeanCounter", "Poisson", "Normal", "RefinedNormal"
    )
  )
  
  # check if 'wts' contains only integers (zeros are allowed)
  if(!is.null(wts) && any(is.na(wts) | wts < 0 | abs(wts - round(wts)) > 1e-07))
    stop("'wts' must contain non-negative integers!")
  
  if(!is.null(wts) && length(wts) != n)
    stop("'probs' and 'wts' (if not NULL) must have the same length!")
  
  ## expand 'probs' according to the counts in 'wts'
  # if 'wts' is NULL, set it to be a vector of ones
  if(is.null(wts)) wts <- rep(1, n)
  
  # expand 'probs'
  probs <- rep(probs, wts)
  
  # re-compute length of 'probs' (= sum of 'wts')
  n <- sum(wts)
  
  # if x = NULL, return all possible probabilities
  if(is.null(x)) x <- 0:n
  
  # identify valid 'x' values (invalid ones will have 0-probability)
  idx.x <- which(x >= 0 & x <= n)
  
  # select valid observations
  y <- x[idx.x]
  
  ## compute probabilities
  # vector for storing the probabilities
  d <- rep(as.numeric(!lower.tail), length(x))
  
  # no computation needed, if there are no valid observations in 'x'
  if(length(idx.x)) {
    # which probabilities are 0 or 1
    idx0  <- which(probs == 0)
    idx1  <- which(probs == 1)
    probs <- probs[probs > 0 & probs < 1]
    
    # number of zeros and ones
    n0 <- length(idx0)
    n1 <- length(idx1)
    np <- n - n0 - n1
    
    # relevant observations
    idx.y <- which(y %in% n1:(n - n0))
    idx.z <- which(y > n - n0)
    
    if(length(idx.y)) {
      z <- y[idx.y] - n1
      
      if(np == 0) {
        # 'probs' contains only zeros and ones => only one possible observation
        d[idx.x][idx.y] <- if(lower.tail) 1 else 0
      } else {
        if(np == 1) {
          # 'probs' contains only one value in (0, 1) => Bernoulli distribution
          d[idx.x][idx.y] <- if(lower.tail) 
            c(1 - probs, 1)[z + 1] else c(probs, 0)[z + 1]
        } else {
          if(all(probs == probs[1])) {
            # all values of 'probs' are equal => standard binomial distribution
            d[idx.x][idx.y] <- pbinom(
              q = z,
              size = np,
              prob = probs[1],
              lower.tail = lower.tail
            )
          } else {
            # otherwise, compute distribution according to 'method'
            d[idx.x][idx.y] <- switch(
              EXPR = method,
              DivideFFT = ppb_dc(z, probs, lower.tail),
              Convolve = ppb_conv(z, probs, lower.tail),
              Characteristic = ppb_dftcf(z, probs, lower.tail),
              Recursive = ppb_rf(z, probs, lower.tail),
              Mean = ppb_mean(z, probs, lower.tail),
              GeoMean = ppb_gmba(z, probs, FALSE, lower.tail),
              GeoMeanCounter = ppb_gmba(z, probs, TRUE, lower.tail),
              Poisson = ppb_pa(z, probs, lower.tail),
              Normal = ppb_na(z, probs, FALSE, lower.tail),
              RefinedNormal = ppb_na(z, probs, TRUE, lower.tail)
            )
          }
        }
      }
    }
    # fill cumulative probabilities of values above the relevant range
    if(length(idx.z)) d[idx.x][idx.z] <- as.double(lower.tail)
  }
  # fill cumulative probabilities of values above n
  d[x > n] <- as.double(lower.tail)
  
  # logarithm, if required
  if(log.p) d <- log(d)
  
  # return results
  return(d)
}

#' @rdname PoissonBinomial-Distribution
#' @importFrom stats stepfun
#' @export
qpbinom <- function(
  p,
  probs,
  wts = NULL,
  method = "DivideFFT",
  lower.tail = TRUE,
  log.p = FALSE
) {
  ## preliminary checks
  # check if 'p' contains only probabilities
  if(!log.p) {
    if(is.null(p) || any(is.na(p) | p < 0 | p > 1))
      stop("'p' must contain real numbers between 0 and 1!")
  } else {
    if(is.null(p) || any(is.na(p) | p > 0))
      stop("'p' must contain real numbers between -Inf and 0!")
  }
  
  # make sure that the value of 'method' matches an implemented procedure
  method <- match.arg(
    arg = method, 
    choices = c(
      "DivideFFT", "Convolve", "Characteristic", "Recursive", "Mean",
      "GeoMean", "GeoMeanCounter", "Poisson", "Normal", "RefinedNormal"
    )
  )
  
  ## compute probabilities (does checking for the other variables)
  cdf <- ppbinom(NULL, probs, wts, method, lower.tail)
  
  # size of distribution
  size <- length(probs)
  
  # logarithm, if required
  if(log.p) p <- exp(p)
  
  ## compute quantiles
  # observable range and indices
  n0 <- sum(probs == 0)
  n1 <- sum(probs == 1)
  hi <- size - n0
  range <- n1:hi
  
  # handle quantiles between 0 and 1
  Q <- if(lower.tail) 
    stepfun(cdf[range + 1], c(range, hi), right = TRUE) else 
      stepfun(rev(cdf[range + 1]), c(hi, rev(range)), right = TRUE)
  
  # vector to store results
  res <- Q(p)
  
  # handle quantiles of 0 or 1
  res[p == lower.tail]  <- hi
  res[p == !lower.tail] <- n1
  
  # return results
  return(res)
}

#' @rdname PoissonBinomial-Distribution
#' @importFrom stats runif rbinom
#' @export
rpbinom <- function(
  n,
  probs,
  wts = NULL,
  method = "DivideFFT",
  generator = "Sample"
) {
  len <- length(n)
  if(len > 1) n <- len
  
  # check if 'n' is NULL
  if(is.null(n)) stop("'n' must not be NULL!")
  
  # number of probabilities
  len <- length(probs)
  
  # check if 'probs' contains only probabilities
  if(is.null(probs) || any(is.na(probs) | probs < 0 | probs > 1))
    stop("'probs' must contain real numbers between 0 and 1!")
  
  # check if 'wts' contains only integers (zeros are allowed)
  if(!is.null(wts) && any(is.na(wts) | wts < 0 | abs(wts - round(wts)) > 1e-07))
    stop("'wts' must contain non-negative integers!")
  
  if(!is.null(wts) && length(wts) != len)
    stop("'probs' and 'wts' (if not NULL) must have the same length!")
  
  ## expand 'probs' according to the counts in 'wts'
  # if 'wts' is NULL, set it to be a vector of ones
  if(is.null(wts)) wts <- rep(1, len)
  
  # expand 'probs'
  probs <- rep(probs, wts)
  
  # make sure that the value of 'method' matches an implemented procedure
  method <- match.arg(
    arg = method, 
    choices = c(
      "DivideFFT", "Convolve", "Characteristic", "Recursive", "Mean",
      "GeoMean", "GeoMeanCounter", "Poisson", "Normal", "RefinedNormal"
    )
  )
  
  # make sure that the value of 'generator' matches an implemented procedure
  generator <- match.arg(generator, c("Sample", "Bernoulli"))
  
  # generate random numbers
  res <- switch(
    EXPR = generator,
    Sample = sample(
      x = 0:length(probs),
      size = n,
      replace = TRUE,
      prob = dpbinom(NULL, probs, NULL, method)
    ),
    Bernoulli = rpb_bernoulli(n, probs)
  )
  
  # return results
  return(res)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#'
NULL

vectorGCD <- function(x) {
    .Call(`_PoissonBinomial_vectorGCD`, x)
}

dpb_conv <- function(obs, probs) {
    .Call(`_PoissonBinomial_dpb_conv`, obs, probs)
}

ppb_conv <- function(obs, probs, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_ppb_conv`, obs, probs, lower_tail)
}

dpb_dc <- function(obs, probs) {
    .Call(`_PoissonBinomial_dpb_dc`, obs, probs)
}

ppb_dc <- function(obs, probs, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_ppb_dc`, obs, probs, lower_tail)
}

dpb_dftcf <- function(obs, probs) {
    .Call(`_PoissonBinomial_dpb_dftcf`, obs, probs)
}

ppb_dftcf <- function(obs, probs, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_ppb_dftcf`, obs, probs, lower_tail)
}

dpb_rf <- function(obs, probs) {
    .Call(`_PoissonBinomial_dpb_rf`, obs, probs)
}

ppb_rf <- function(obs, probs, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_ppb_rf`, obs, probs, lower_tail)
}

dpb_mean <- function(obs, probs) {
    .Call(`_PoissonBinomial_dpb_mean`, obs, probs)
}

ppb_mean <- function(obs, probs, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_ppb_mean`, obs, probs, lower_tail)
}

dpb_gmba <- function(obs, probs, anti = FALSE) {
    .Call(`_PoissonBinomial_dpb_gmba`, obs, probs, anti)
}

ppb_gmba <- function(obs, probs, anti = FALSE, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_ppb_gmba`, obs, probs, anti, lower_tail)
}

dpb_pa <- function(obs, probs) {
    .Call(`_PoissonBinomial_dpb_pa`, obs, probs)
}

ppb_pa <- function(obs, probs, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_ppb_pa`, obs, probs, lower_tail)
}

ppb_na <- function(obs, probs, refined = TRUE, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_ppb_na`, obs, probs, refined, lower_tail)
}

dpb_na <- function(obs, probs, refined = TRUE) {
    .Call(`_PoissonBinomial_dpb_na`, obs, probs, refined)
}

rpb_bernoulli <- function(n, probs) {
    .Call(`_PoissonBinomial_rpb_bernoulli`, n, probs)
}

dgpb_conv <- function(obs, probs, val_p, val_q) {
    .Call(`_PoissonBinomial_dgpb_conv`, obs, probs, val_p, val_q)
}

pgpb_conv <- function(obs, probs, val_p, val_q, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_pgpb_conv`, obs, probs, val_p, val_q, lower_tail)
}

dgpb_dc <- function(obs, probs, val_p, val_q) {
    .Call(`_PoissonBinomial_dgpb_dc`, obs, probs, val_p, val_q)
}

pgpb_dc <- function(obs, probs, val_p, val_q, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_pgpb_dc`, obs, probs, val_p, val_q, lower_tail)
}

dgpb_dftcf <- function(obs, probs, val_p, val_q) {
    .Call(`_PoissonBinomial_dgpb_dftcf`, obs, probs, val_p, val_q)
}

pgpb_dftcf <- function(obs, probs, val_p, val_q, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_pgpb_dftcf`, obs, probs, val_p, val_q, lower_tail)
}

pgpb_na <- function(obs, probs, val_p, val_q, refined = TRUE, lower_tail = TRUE) {
    .Call(`_PoissonBinomial_pgpb_na`, obs, probs, val_p, val_q, refined, lower_tail)
}

dgpb_na <- function(obs, probs, val_p, val_q, refined = TRUE) {
    .Call(`_PoissonBinomial_dgpb_na`, obs, probs, val_p, val_q, refined)
}

rgpb_bernoulli <- function(n, probs, val_p, val_q) {
    .Call(`_PoissonBinomial_rgpb_bernoulli`, n, probs, val_p, val_q)
}

# Register entry points for exported C++ functions
methods::setLoadAction(function(ns) {
    .Call(`_PoissonBinomial_RcppExport_registerCCallable`)
})

check.args.GPB <- function(x, probs, val_p, val_q, wts, method, log.p = FALSE){
  # check if 'x' contains only integers
  if(!is.null(x) && any(x - round(x) != 0)){
    warning("'x' should contain integers only! Using rounded off values.")
    x <- floor(x)
  }
  
  # check if 'probs' contains only probabilities
  if(is.null(probs) || any(is.na(probs) | probs < 0 | probs > 1))
    stop("'probs' must contain real numbers between 0 and 1!")
  
  # number of probabilities
  n <- length(probs)
  
  # check if 'val_p' and 'val_q' have the same length as 'probs'
  if(length(val_p) != n || length(val_q) != n) stop("'probs', 'val_p' and 'val_q' must have the same length!")
  
  if(!is.null(wts) && length(wts) != n)
    stop("'probs' and 'wts' (if not NULL) must have the same length!")
  
  # check if 'val_p' contains only integers
  if(!is.null(val_p) && any(val_p - round(val_p) != 0)){
    warning("'val_p' should contain integers only! Using rounded off values.")
    val_p <- floor(val_p)
  }
  
  # check if 'val_q' contains only integers
  if(!is.null(val_q) && any(val_q - round(val_q) != 0)){
    warning("'val_q' should contain integers only! Using rounded off values.")
    val_q <- floor(val_q)
  }
  
  # check if 'wts' contains only integers (zeros are allowed)
  if(!is.null(wts) && any(is.na(wts) | wts < 0 | abs(wts - round(wts)) > 1e-07))
    stop("'wts' must contain non-negative integers!")
  
  # make sure that the value of 'method' matches one of the implemented procedures
  method <- match.arg(method, c("DivideFFT", "Convolve", "Characteristic", "Normal", "RefinedNormal"))
  
  # if all checks were successful, return matched 'method'
  return(method)
}


transform.GPB <- function(x, probs, val_p, val_q, wts){
  # number of probabilities
  n <- length(probs)
  
  ## expand 'probs', 'val_p' and 'val_q' according to the counts in 'wts'
  # if 'wts' is NULL, set it to be a vector of ones
  if(is.null(wts))
    wts <- rep(1, n)
  
  # expand 'probs', 'val_p', 'val_q'
  probs <- rep(probs, wts)
  val_p <- rep(val_p, wts)
  val_q <- rep(val_q, wts)
  
  # reorder 'val_p' and 'val_q' so that values in 'val_p' are always greater
  val_gr <- pmax(val_p, val_q)
  val_lo <- pmin(val_p, val_q)
  probs[val_gr > val_p] <- 1 - probs[val_gr > val_p]
  
  # re-compute length of 'probs' (= sum of 'wts')
  n <- sum(wts)
  
  ## determine relevant range of observations
  # determine minimum and maximum possible observations
  sum_min <- sum(val_lo)
  sum_max <- sum(val_gr)
  
  # which probabilities are 0 or 1, which val_p and val_q are equal
  idx.0 <- which(probs == 0)
  idx.1 <- which(probs == 1)
  idx.v <- which(val_gr == val_lo & probs > 0 & probs < 1)
  idx.r <- setdiff(1:n, union(union(idx.0, idx.1), idx.v))
  
  # guaranteed
  val_gr_sure <- val_gr[idx.1]
  val_lo_sure <- val_lo[idx.0]
  vals_equal <- val_gr[idx.v]# equal to val_lo[idx.v]
  sum_sure <- sum(val_gr_sure, val_lo_sure, vals_equal)
  
  # limit 'probs', 'val_p' and 'val_q' to relevant range
  np <- length(idx.r)
  if(np){
    probs <- probs[idx.r]
    val_gr <- val_gr[idx.r]
    val_lo <- val_lo[idx.r]
  }else{
    probs <- 1
    val_gr <- 0
    val_lo <- 0
  }
  
  # compute differences and their GCD
  diffs <- val_gr - val_lo
  
  # bounds of relevant observations
  sum_min_in <- sum(val_lo) + sum_sure
  sum_max_in <- sum(val_gr) + sum_sure
  
  return(list(probs = probs, val_p = val_gr, val_q = val_lo, compl.range = sum_min:sum_max, inner.range = sum_min_in:sum_max_in, inner.size = sum_max_in - sum_min_in + 1, n = np, diffs = diffs))
}

#' @name PoissonBinomial-package
#'
#' @title
#' Efficient Exact and Approximate Implementations for Computing Ordinary and
#' Generalised Poisson Binomial Distributions
#'
#' @description
#' This package implements various algorithms for computing the probability mass
#' function, the cumulative distribution function, quantiles and random numbers
#' of both ordinary and generalised Poisson binomial distributions.
#'
#' @import Rcpp
#' @useDynLib PoissonBinomial, .registration = TRUE
#'
#' @section References:
#' Hong, Y. (2013). On computing the distribution function for the Poisson
#'   binomial distribution. \emph{Computational Statistics & Data Analysis},
#'   \strong{59}, pp. 41-51. \doi{10.1016/j.csda.2012.10.006}
#'
#' Biscarri, W., Zhao, S. D. and Brunner, R. J. (2018) A simple and fast method
#'   for computing the Poisson binomial distribution.
#'   \emph{Computational Statistics and Data Analysis}, \strong{31}, pp.
#'   216–222. \doi{10.1016/j.csda.2018.01.007}
#'    
#' Zhang, M., Hong, Y. and Balakrishnan, N. (2018). The generalized 
#'   Poisson-binomial distribution and the computation of its distribution
#'   function. \emph{Journal of Statistical Computational and Simulation},
#'   \strong{88}(8), pp. 1515-1527. \doi{10.1080/00949655.2018.1440294}
#'    
#' @examples
#' # Functions for ordinary Poisson binomial distributions
#' set.seed(1)
#' pp <- c(1, 0, runif(10), 1, 0, 1)
#' qq <- seq(0, 1, 0.01)
#'
#' dpbinom(NULL, pp)
#' ppbinom(7:10, pp, method = "DivideFFT")
#' qpbinom(qq, pp, method = "Convolve")
#' rpbinom(10, pp, method = "RefinedNormal")
#'
#' # Functions for generalised Poisson binomial distributions
#' va <- rep(5, length(pp))
#' vb <- 1:length(pp)
#'
#' dgpbinom(NULL, pp, va, vb, method = "Convolve")
#' pgpbinom(80:100, pp, va, vb, method = "Convolve")
#' qgpbinom(qq, pp, va, vb, method = "Convolve")
#' rgpbinom(100, pp, va, vb, method = "Convolve")
"_PACKAGE"

#' @name GenPoissonBinomial-Distribution
#'
#' @title The Generalised Poisson Binomial Distribution
#'
#' @description
#' Density, distribution function, quantile function and random generation for
#' the generalised Poisson binomial distribution with probability vector
#' \code{probs}.
#'
#' @param x           Either a vector of observed sums or NULL. If NULL,
#'                    probabilities of all possible observations are
#'                    returned.
#' @param p           Vector of probabilities for computation of quantiles.
#' @param n           Number of observations. If \code{length(n) > 1}, the
#'                    length is taken to be the number required.
#' @param probs       Vector of probabilities of success of each Bernoulli
#'                    trial.
#' @param val_p       Vector of values that each trial produces with probability
#'                    in \code{probs}.
#' @param val_q       Vector of values that each trial produces with probability
#'                    in \code{1 - probs}.
#' @param method      Character string that specifies the method of computation
#'                    and must be one of \code{"DivideFFT"}, \code{"Convolve"}, 
#'                    \code{"Characteristic"}, \code{"Normal"} or
#'                    \code{"RefinedNormal"} (abbreviations are allowed).
#' @param wts         Vector of non-negative integer weights for the input
#'                    probabilities.
#' @param log,log.p   Logical value indicating if results are given as
#'                    logarithms.
#' @param lower.tail  Logical value indicating if results are \eqn{P[X \leq x]}
#'                    (if \code{TRUE}; default) or \eqn{P[X > x]} (if 
#'                    \code{FALSE}).
#' @param generator   Character string that specifies the random number
#'                    generator and must either be \code{"Sample"} or
#'                    \code{"Bernoulli"} (abbreviations are allowed).
#'
#' @details
#' See the references for computational details. The \emph{Divide and Conquer}
#' (\code{"DivideFFT"}) and \emph{Direct Convolution} (\code{"Convolve"})
#' algorithms are derived and described in Biscarri, Zhao & Brunner (2018). They
#' have been modified for use with the generalised Poisson binomial
#' distribution. The
#' \emph{Discrete Fourier Transformation of the Characteristic Function}
#' (\code{"Characteristic"}) is derived in Zhang, Hong & Balakrishnan (2018),
#' the \emph{Normal Approach} (\code{"Normal"}) and the
#' \emph{Refined Normal Approach} (\code{"RefinedNormal"}) are described in Hong
#' (2013). They were slightly adapted for the generalised Poisson binomial
#' distribution.
#'
#' In some special cases regarding the values of \code{probs}, the \code{method}
#' parameter is ignored (see Introduction vignette).
#'
#' Random numbers can be generated in two ways. The \code{"Sample"} method
#' uses \code{R}'s \code{sample} function to draw random values according to
#' their probabilities that are calculated by \code{dgpbinom}. The
#' \code{"Bernoulli"} procedure ignores the \code{method} parameter and
#' simulates Bernoulli-distributed random numbers according to the probabilities
#' in \code{probs} and sums them up. It is a bit slower than the \code{"Sample"}
#' generator, but may yield better results, as it allows to obtain observations
#' that cannot be generated by the \code{"Sample"} procedure, because
#' \code{dgpbinom} may compute 0-probabilities, due to rounding, if the length
#' of \code{probs} is large and/or its values contain a lot of very small
#' values.
#'
#' @return
#' \code{dgpbinom} gives the density, \code{pgpbinom} computes the distribution
#' function, \code{qgpbinom} gives the quantile function and \code{rgpbinom}
#' generates random deviates.
#'
#' For \code{rgpbinom}, the length of the result is determined by \code{n}, and
#' is the lengths of the numerical arguments for the other functions.
#'
#' @references
#' Hong, Y. (2018). On computing the distribution function for the Poisson
#'  binomial distribution. \emph{Computational Statistics & Data Analysis},
#'  \strong{59}, pp. 41-51. \doi{10.1016/j.csda.2012.10.006}
#'
#' Biscarri, W., Zhao, S. D. and Brunner, R. J. (2018) A simple and fast method
#'   for computing the Poisson binomial distribution.
#'   \emph{Computational Statistics and Data Analysis}, \strong{31}, pp.
#'   216–222. \doi{10.1016/j.csda.2018.01.007}
#'    
#' Zhang, M., Hong, Y. and Balakrishnan, N. (2018). The generalized 
#'   Poisson-binomial distribution and the computation of its distribution
#'   function. \emph{Journal of Statistical Computational and Simulation},
#'   \strong{88}(8), pp. 1515-1527. \doi{10.1080/00949655.2018.1440294}
#'    
#' @examples
#' set.seed(1)
#' pp <- c(1, 0, runif(10), 1, 0, 1)
#' qq <- seq(0, 1, 0.01)
#' va <- rep(5, length(pp))
#' vb <- 1:length(pp)
#'
#' dgpbinom(NULL, pp, va, vb, method = "DivideFFT")
#' pgpbinom(75:100, pp, va, vb, method = "DivideFFT")
#' qgpbinom(qq, pp, va, vb, method = "DivideFFT")
#' rgpbinom(100, pp, va, vb, method = "DivideFFT")
#' 
#' dgpbinom(NULL, pp, va, vb, method = "Convolve")
#' pgpbinom(75:100, pp, va, vb, method = "Convolve")
#' qgpbinom(qq, pp, va, vb, method = "Convolve")
#' rgpbinom(100, pp, va, vb, method = "Convolve")
#' 
#' dgpbinom(NULL, pp, va, vb, method = "Characteristic")
#' pgpbinom(75:100, pp, va, vb, method = "Characteristic")
#' qgpbinom(qq, pp, va, vb, method = "Characteristic")
#' rgpbinom(100, pp, va, vb, method = "Characteristic")
#'
#' dgpbinom(NULL, pp, va, vb, method = "Normal")
#' pgpbinom(75:100, pp, va, vb, method = "Normal")
#' qgpbinom(qq, pp, va, vb, method = "Normal")
#' rgpbinom(100, pp, va, vb, method = "Normal")
#' 
#' dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
#' pgpbinom(75:100, pp, va, vb, method = "RefinedNormal")
#' qgpbinom(qq, pp, va, vb, method = "RefinedNormal")
#' rgpbinom(100, pp, va, vb, method = "RefinedNormal")
#' 
#' @export
dgpbinom <- function(
  x,
  probs,
  val_p,
  val_q,
  wts = NULL,
  method = "DivideFFT",
  log = FALSE
) {
  ## preliminary checks
  method <- check.args.GPB(x, probs, val_p, val_q, wts, method)
  
  ## transform input to relevant range
  transf <- transformGPB(x, probs, val_p, val_q, wts)
  
  # if x = NULL, return all possible probabilities
  if(is.null(x)) x <- transf$compl.range
  
  # identify valid 'x' values (invalid ones will have 0-probability)
  idx.valid <- which(x %in% transf$compl.range)
  
  ## compute probabilities
  # vector for storing the probabilities
  d <- double(length(x))
  
  # no computation needed, if there are no valid observations in 'x'
  if(length(idx.valid)) {
    # select valid observations in relevant range
    y <- x[idx.valid]
    
    # relevant observations
    idx.inner <- which(y %in% transf$inner.range)
    
    # if input values are not in relevant range, they are impossible (i.e.
    #  return 0-probabilities)
    if(length(idx.inner)) {
      # transformed input parameters
      n <- transf$n
      probs <- transf$probs
      diffs <- transf$diffs
      
      if(n == 0) {
        # 'probs' contains only zeros and ones => only one possible observation
        d[idx.valid][idx.inner] <- 1
      } else {
        z <- y[idx.inner] - transf$inner.range[1]
        # compute distribution
        if(all(diffs == diffs[1])) {
          # all values of 'diffs' are equal => scaled ordinary PBD
          remainder <- z %% diffs[1]
          idx.r <- which(remainder == 0)
          d[idx.valid][idx.inner][idx.r] <- dpbinom(
            x = (z %/% diffs[1])[idx.r], 
            probs = probs,
            method = method
          )
        } else {
          # compute distribution according to 'method'
          d[idx.valid][idx.inner] <- switch(
            EXPR = method,
            DivideFFT = dgpb_dc(z, probs, diffs, rep(0, n)),
            Convolve = dgpb_conv(z, probs, diffs, rep(0, n)),
            Characteristic = dgpb_dftcf(z, probs, diffs, rep(0, n)),
            Normal = dgpb_na(z, probs, diffs, rep(0, n), FALSE),
            RefinedNormal = dgpb_na(z, probs, diffs, rep(0, n), TRUE)
          )
        }
      }
    }
  }
  
  # logarithm, if required
  if(log) d <- log(d)
  
  # return results
  return(d)
}

#' @rdname GenPoissonBinomial-Distribution
#' @export
pgpbinom <- function(
  x,
  probs,
  val_p,
  val_q,
  wts = NULL,
  method = "DivideFFT",
  lower.tail = TRUE,
  log.p = FALSE
) {
  ## preliminary checks
  method <- check.args.GPB(x, probs, val_p, val_q, wts, method)
  
  ## transform input to relevant range
  transf <- transformGPB(x, probs, val_p, val_q, wts)
  
  # if x = NULL, return all possible probabilities
  if(is.null(x)) x <- transf$compl.range
  
  # identify valid 'x' values (invalid ones will have 0-probability)
  idx.valid <- which(x %in% transf$compl.range)
  
  ## compute probabilities
  # vector for storing the probabilities
  d <- rep(as.numeric(!lower.tail), length(x))
  
  # no computation needed, if there are no valid observations in 'x'
  if(length(idx.valid)) {
    # select valid observations in relevant range
    y <- x[idx.valid]
    
    # relevant observations
    idx.inner <- which(y %in% transf$inner.range)
    
    if(length(idx.inner)) {
      # transformed input parameters
      n <- transf$n
      probs <- transf$probs
      diffs <- transf$diffs
      
      if(n == 0) {
        # 'probs' contains only zeros and ones => only one possible observation
        d[idx.valid][idx.inner] <- as.numeric(lower.tail)
      } else {
        # select and rescale relevant observations
        z <- y[idx.inner] - transf$inner.range[1]
        
        # compute distribution
        if(all(diffs == diffs[1])) {
          # all GCD-optimised values are equal => standard binomial distribution
          d[idx.valid][idx.inner] <- ppbinom(
            x = z %/% diffs[1],
            probs = probs,
            method = method,
            lower.tail = lower.tail
          )
        } else {
          # compute distribution according to 'method'
          d[idx.valid][idx.inner] <- switch(
            EXPR = method,
            DivideFFT = pgpb_dc(z, probs, diffs, rep(0, n), lower.tail),
            Convolve = pgpb_conv(z, probs, diffs, rep(0, n), lower.tail),
            Characteristic = pgpb_dftcf(z, probs, diffs, rep(0, n), lower.tail),
            Normal = pgpb_na(z, probs, diffs, rep(0, n), FALSE, lower.tail),
            RefinedNormal = pgpb_na(z, probs, diffs, rep(0, n), TRUE, lower.tail)
          )
        }
      }
    }
    
    # which valid observations are above relevant range
    idx.above <- which(y > max(transf$inner.range))
    
    # fill cumulative probabilities of values above the relevant range
    if(length(idx.above)) d[idx.valid][idx.above] <- as.double(lower.tail)
  }
  
  # fill cumulative probabilities of values above complete range
  d[x > max(transf$compl.range)] <- as.double(lower.tail)
  
  # logarithm, if required
  if(log.p) d <- log(d)
  
  # return results
  return(d)
}

#' @rdname GenPoissonBinomial-Distribution
#' @importFrom stats stepfun
#' @export
qgpbinom <- function(
  p,
  probs,
  val_p,
  val_q,
  wts = NULL,
  method = "DivideFFT",
  lower.tail = TRUE,
  log.p = FALSE
) {
  ## preliminary checks
  method <- check.args.GPB(NULL, probs, val_p, val_q, wts, method)
  
  # check if 'q' contains only probabilities
  if(!log.p) {
    if(is.null(p) || any(is.na(p) | p < 0 | p > 1))
      stop("'p' must contain real numbers between 0 and 1!")
  } else {
    if(is.null(p) || any(is.na(p) | p > 0))
      stop("'p' must contain real numbers between -Inf and 0!")
  }
  
  ## transform input to relevant range
  transf <- transformGPB(NULL, probs, val_p, val_q, wts)
  probs  <- transf$probs
  val_p  <- transf$val_p
  val_q  <- transf$val_q
  
  ## compute probabilities (does checking for the other variables)
  cdf <- pgpbinom(NULL, probs, val_p, val_q, NULL, method, lower.tail)
  
  # bounds of relevant observations
  first <- min(transf$inner.range)
  last  <- max(transf$inner.range)
  
  # length of cdf
  len <- length(cdf)
  
  # logarithm, if required
  if(log.p) p <- exp(p)
  
  ## compute quantiles
  # handle quantiles between 0 and 1
  Q <- if(lower.tail) {
    stepfun(
      x = cdf[transf$inner.range - first + 1],
      y = c(transf$inner.range, last),
      right = TRUE
    )
  } else {
    stepfun(
      x = rev(cdf[transf$inner.range - first + 1]),
      y = c(last, rev(transf$inner.range)),
      right = TRUE
    )
  }
  
  # vector to store results
  res <- Q(p)
  
  # handle quantiles of 0 or 1
  res[p == lower.tail]  <- last
  res[p == !lower.tail] <- first
  
  # return results
  return(res)
}

#' @rdname GenPoissonBinomial-Distribution
#' @importFrom stats runif rbinom
#' @export
rgpbinom <- function(
  n,
  probs,
  val_p,
  val_q,
  wts = NULL,
  method = "DivideFFT",
  generator = "Sample"
) {
  ## preliminary checks
  method <- check.args.GPB(NULL, probs, val_p, val_q, wts, method)
  
  len <- length(n)
  if(len > 1) n <- len
  
  # check if 'n' is NULL
  if(is.null(n)) stop("'n' must not be NULL!")
  
  ## expand 'probs', 'val_p' and 'val_q' according to the counts in 'wts'
  # if 'wts' is NULL, set it to be a vector of ones
  if(is.null(wts)) wts <- rep(1, length(probs))
  
  # expand 'probs', 'val_p', 'val_q'
  probs <- rep(probs, wts)
  val_p <- rep(val_p, wts)
  val_q <- rep(val_q, wts)
  
  # make sure that the value of 'generator' matches one of the implemented procedures
  generator <- match.arg(generator, c("Sample", "Bernoulli"))
  
  # generate random numbers
  res <- switch(
    EXPR = generator,
    Sample = sample(
      x = sum(pmin(val_p, val_q)):sum(pmax(val_p, val_q)),
      size = n,
      replace = TRUE,
      prob = dgpbinom(NULL, probs, val_p, val_q, NULL, method)
    ),
    Bernoulli = rgpb_bernoulli(n, probs, val_p, val_q)
  )
  
  # return results
  return(res)
}

## Test environments
* local Windows 10 Education 22H2 10.0.19045.3693, R 4.3.2
* win-builder (configurations: "release", "oldrelease", "devel")
* R-hub (configurations: "sanitizers", "valgrind", "CRAN")


## R CMD check results

### local
0 errors | 0 warnings | 0 notes

### R-hub
* checking CRAN incoming feasibility ... [12s] NOTE
Maintainer: 'Florian Junge <florian.junge@mailbox.org>'

New maintainer:
  Florian Junge <florian.junge@mailbox.org>
Old maintainer(s):
  Florian Junge <florian.junge@h-da.de>
  
This change of addresses has been announced via eMail to
<cran-submissions@r-project.org> in advance.


### win-builder 
* checking CRAN incoming feasibility ... [12s] NOTE
Maintainer: 'Florian Junge <florian.junge@mailbox.org>'

New maintainer:
  Florian Junge <florian.junge@mailbox.org>
Old maintainer(s):
  Florian Junge <florian.junge@h-da.de>
  
This change of addresses has been announced via eMail to
<cran-submissions@r-project.org> in advance.

---
title: "Approximate Procedures"
output:
  rmarkdown::html_vignette:
    toc: yes
vignette: >
  %\VignetteIndexEntry{Approximate Procedures}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo = FALSE}
library(PoissonBinomial)
```

## Ordinary Poisson Binomial Distribution

### Poisson Approximation

The *Poisson Approximation* (DC) approach is requested with `method = "Poisson"`. It is based on a Poisson distribution, whose parameter is the sum of the probabilities of success.

```{r pa1}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Poisson")
ppbinom(NULL, pp, wt, "Poisson")
```

A comparison with exact computation shows that the approximation quality of the PA procedure increases with smaller probabilities of success. The reason is that the Poisson Binomial distribution approaches a Poisson distribution when the probabilities are very small.

```{r pa2}
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "Poisson")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Poisson") - dpbinom(NULL, pp))

# U(0, 0.01) random probabilities of success
pp <- runif(20, 0, 0.01)
dpbinom(NULL, pp, method = "Poisson")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Poisson") - dpbinom(NULL, pp))
```

### Arithmetic Mean Binomial Approximation

The *Arithmetic Mean Binomial Approximation* (AMBA) approach is requested with `method = "Mean"`. It is based on a Binomial distribution, whose parameter is the arithmetic mean of the probabilities of success.

```{r am1}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
mean(rep(pp, wt))

dpbinom(NULL, pp, wt, "Mean")
ppbinom(NULL, pp, wt, "Mean")
```

A comparison with exact computation shows that the approximation quality of the AMBA procedure increases when the probabilities of success are closer to each other. The reason is that, although the expectation remains unchanged, the distribution's variance becomes smaller the less the probabilities differ. Since this variance is minimized by equal probabilities (but still underestimated), the AMBA method is best suited for situations with very similar probabilities of success.

```{r am2}
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "Mean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Mean") - dpbinom(NULL, pp))

# U(0.3, 0.5) random probabilities of success
pp <- runif(20, 0.3, 0.5)
dpbinom(NULL, pp, method = "Mean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Mean") - dpbinom(NULL, pp))

# U(0.39, 0.41) random probabilities of success
pp <- runif(20, 0.39, 0.41)
dpbinom(NULL, pp, method = "Mean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Mean") - dpbinom(NULL, pp))
```

### Geometric Mean Binomial Approximation - Variant A

The *Geometric Mean Binomial Approximation (Variant A)* (GMBA-A) approach is requested with `method = "GeoMean"`. It is based on a Binomial distribution, whose parameter is the geometric mean of the probabilities of success:
$$\hat{p} = \sqrt[n]{p_1 \cdot ... \cdot p_n}$$

```{r gma1}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
prod(rep(pp, wt))^(1/sum(wt))

dpbinom(NULL, pp, wt, "GeoMean")
ppbinom(NULL, pp, wt, "GeoMean")
```

It is known that the geometric mean of the probabilities of success is always smaller than their arithmetic mean. Thus, we get a stochastically *smaller* binomial distribution. A comparison with exact computation shows that the approximation quality of the GMBA-A procedure increases when the probabilities of success are closer to each other:

```{r gma2}
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "GeoMean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMean") - dpbinom(NULL, pp))

# U(0.4, 0.6) random probabilities of success
pp <- runif(20, 0.4, 0.6)
dpbinom(NULL, pp, method = "GeoMean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMean") - dpbinom(NULL, pp))

# U(0.49, 0.51) random probabilities of success
pp <- runif(20, 0.49, 0.51)
dpbinom(NULL, pp, method = "GeoMean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMean") - dpbinom(NULL, pp))
```

### Geometric Mean Binomial Approximation - Variant B

The *Geometric Mean Binomial Approximation (Variant B)* (GMBA-B) approach is requested with `method = "GeoMeanCounter"`. It is based on a Binomial distribution, whose parameter is 1 minus the geometric mean of the probabilities of **failure**:
$$\hat{p} = 1 - \sqrt[n]{(1 - p_1) \cdot ... \cdot (1 - p_n)}$$

```{r gmb1}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
1 - prod(1 - rep(pp, wt))^(1/sum(wt))

dpbinom(NULL, pp, wt, "GeoMeanCounter")
ppbinom(NULL, pp, wt, "GeoMeanCounter")
```

It is known that the geometric mean of the probabilities of **failure** is always smaller than their arithmetic mean. As a result, 1 minus the geometric mean is larger than 1 minus the arithmetic mean. Thus, we get a stochastically *larger* binomial distribution. A comparison with exact computation shows that the approximation quality of the GMBA-B procedure again increases when the probabilities of success are closer to each other:

```{r gmb2}
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "GeoMeanCounter")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMeanCounter") - dpbinom(NULL, pp))

# U(0.4, 0.6) random probabilities of success
pp <- runif(20, 0.4, 0.6)
dpbinom(NULL, pp, method = "GeoMeanCounter")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMeanCounter") - dpbinom(NULL, pp))

# U(0.49, 0.51) random probabilities of success
pp <- runif(20, 0.49, 0.51)
dpbinom(NULL, pp, method = "GeoMeanCounter")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMeanCounter") - dpbinom(NULL, pp))
```

### Normal Approximation

The *Normal Approximation* (NA) approach is requested with `method = "Normal"`. It is based on a Normal distribution, whose parameters are derived from the theoretical mean and variance of the input probabilities of success.

```{r na1-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Normal")
ppbinom(NULL, pp, wt, "Normal")
```

A comparison with exact computation shows that the approximation quality of the NA procedure increases with larger numbers of probabilities of success:

```{r na2-ord}
set.seed(1)

# 10 random probabilities of success
pp <- runif(10)
dpn <- dpbinom(NULL, pp, method = "Normal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
dpn <- dpbinom(NULL, pp, method = "Normal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100000 random probabilities of success
pp <- runif(100000)
dpn <- dpbinom(NULL, pp, method = "Normal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])
```

### Refined Normal Approximation

The *Refined Normal Approximation* (RNA) approach is requested with `method = "RefinedNormal"`. It is based on a Normal distribution, whose parameters are derived from the theoretical mean, variance and skewness of the input probabilities of success.

```{r rna1-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "RefinedNormal")
ppbinom(NULL, pp, wt, "RefinedNormal")
```

A comparison with exact computation shows that the approximation quality of the RNA procedure increases with larger numbers of probabilities of success:

```{r rna2-ord}
set.seed(1)

# 10 random probabilities of success
pp <- runif(10)
dpn <- dpbinom(NULL, pp, method = "RefinedNormal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
dpn <- dpbinom(NULL, pp, method = "RefinedNormal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100000 random probabilities of success
pp <- runif(100000)
dpn <- dpbinom(NULL, pp, method = "RefinedNormal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])
```

### Processing Speed Comparisons

To assess the performance of the approximation procedures, we use the `microbenchmark` package. Each algorithm has to calculate the PMF repeatedly based on random probability vectors. The run times are then summarized in a table that presents, among other statistics, their minima, maxima and means. The following results were recorded on an AMD Ryzen 9 5900X with 64 GiB of RAM and Manjaro KDE Linux (26.0.2).

```{r benchmark-ord}
library(microbenchmark)
set.seed(1)

f1 <- function() dpbinom(NULL, runif(4000), method = "Normal")
f2 <- function() dpbinom(NULL, runif(4000), method = "Poisson")
f3 <- function() dpbinom(NULL, runif(4000), method = "RefinedNormal")
f4 <- function() dpbinom(NULL, runif(4000), method = "Mean")
f5 <- function() dpbinom(NULL, runif(4000), method = "GeoMean")
f6 <- function() dpbinom(NULL, runif(4000), method = "GeoMeanCounter")
f7 <- function() dpbinom(NULL, runif(4000), method = "DivideFFT")

microbenchmark(f1(), f2(), f3(), f4(), f5(), f6(), f7(), times = 51)
```

Clearly, the NA procedure is the fastest, followed by the PA and RNA methods. The next fastest algorithms are AMBA, GMBA-A and GMBA-B. They exhibit almost equal mean execution speed, with the AMBA algorithm being slightly faster. All of the approximation procedures outperform the fastest exact approach, DC-FFT, by far.


## Generalized Poisson Binomial Distribution

### Generalized Normal Approximation

The *Generalized Normal Approximation* (G-NA) approach is requested with `method = "Normal"`. It is based on a Normal distribution, whose parameters are derived from the theoretical mean and variance of the input probabilities of success (see [Introduction](intro.html).

```{r na1-gen}
set.seed(2)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)

dgpbinom(NULL, pp, va, vb, wt, "Normal")
pgpbinom(NULL, pp, va, vb, wt, "Normal")
```

A comparison with exact computation shows that the approximation quality of the NA procedure increases with larger numbers of probabilities of success:

```{r na2-gen}
set.seed(2)

# 10 random probabilities of success
pp <- runif(10)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "Normal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100 random probabilities of success
pp <- runif(100)
va <- sample(0:100, 100, TRUE)
vb <- sample(0:100, 100, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "Normal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
va <- sample(0:1000, 1000, TRUE)
vb <- sample(0:1000, 1000, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "Normal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])
```

### Generalized Refined Normal Approximation

The *Generalized Refined Normal Approximation* (G-RNA) approach is requested with `method = "RefinedNormal"`. It is based on a Normal distribution, whose parameters are derived from the theoretical mean, variance and skewness of the input probabilities of success.

```{r rna1-gen}
set.seed(2)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)
dgpbinom(NULL, pp, va, vb, wt, "RefinedNormal")
pgpbinom(NULL, pp, va, vb, wt, "RefinedNormal")
```

A comparison with exact computation shows that the approximation quality of the RNA procedure increases with larger numbers of probabilities of success:

```{r rna2-gen}
set.seed(2)

# 10 random probabilities of success
pp <- runif(10)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100 random probabilities of success
pp <- runif(100)
va <- sample(0:100, 100, TRUE)
vb <- sample(0:100, 100, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
va <- sample(0:1000, 1000, TRUE)
vb <- sample(0:1000, 1000, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])
```

### Processing Speed Comparisons

To assess the performance of the approximation procedures, we use the `microbenchmark` package. Each algorithm has to calculate the PMF repeatedly based on random probability vectors. The run times are then summarized in a table that presents, among other statistics, their minima, maxima and means. The following results were recorded on an AMD Ryzen 9 5900X with 64 GiB of RAM and Manjaro KDE Linux (26.0.2).

```{r benchmark-gen}
library(microbenchmark)
n <- 1500
set.seed(2)
va <- sample(1:50, n, TRUE)
vb <- sample(1:50, n, TRUE)

f1 <- function() dgpbinom(NULL, runif(n), va, vb, method = "Normal")
f2 <- function() dgpbinom(NULL, runif(n), va, vb, method = "RefinedNormal")
f3 <- function() dgpbinom(NULL, runif(n), va, vb, method = "DivideFFT")

microbenchmark(f1(), f2(), f3(), times = 51)
```

Clearly, the G-NA procedure is the fastest, followed by the G-RNA method. Both are hugely faster than G-DC-FFT.

---
title: "Approximate Procedures"
output:
  rmarkdown::html_vignette:
    toc: yes
vignette: >
  %\VignetteIndexEntry{Approximate Procedures}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo = FALSE}
library(PoissonBinomial)
```

## Ordinary Poisson Binomial Distribution

### Poisson Approximation

The *Poisson Approximation* (DC) approach is requested with `method = "Poisson"`. It is based on a Poisson distribution, whose parameter is the sum of the probabilities of success.

```{r pa1}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Poisson")
ppbinom(NULL, pp, wt, "Poisson")
```

A comparison with exact computation shows that the approximation quality of the PA procedure increases with smaller probabilities of success. The reason is that the Poisson Binomial distribution approaches a Poisson distribution when the probabilities are very small.

```{r pa2}
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "Poisson")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Poisson") - dpbinom(NULL, pp))

# U(0, 0.01) random probabilities of success
pp <- runif(20, 0, 0.01)
dpbinom(NULL, pp, method = "Poisson")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Poisson") - dpbinom(NULL, pp))
```

### Arithmetic Mean Binomial Approximation

The *Arithmetic Mean Binomial Approximation* (AMBA) approach is requested with `method = "Mean"`. It is based on a Binomial distribution, whose parameter is the arithmetic mean of the probabilities of success.

```{r am1}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
mean(rep(pp, wt))

dpbinom(NULL, pp, wt, "Mean")
ppbinom(NULL, pp, wt, "Mean")
```

A comparison with exact computation shows that the approximation quality of the AMBA procedure increases when the probabilities of success are closer to each other. The reason is that, although the expectation remains unchanged, the distribution's variance becomes smaller the less the probabilities differ. Since this variance is minimized by equal probabilities (but still underestimated), the AMBA method is best suited for situations with very similar probabilities of success.

```{r am2}
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "Mean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Mean") - dpbinom(NULL, pp))

# U(0.3, 0.5) random probabilities of success
pp <- runif(20, 0.3, 0.5)
dpbinom(NULL, pp, method = "Mean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Mean") - dpbinom(NULL, pp))

# U(0.39, 0.41) random probabilities of success
pp <- runif(20, 0.39, 0.41)
dpbinom(NULL, pp, method = "Mean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "Mean") - dpbinom(NULL, pp))
```

### Geometric Mean Binomial Approximation - Variant A

The *Geometric Mean Binomial Approximation (Variant A)* (GMBA-A) approach is requested with `method = "GeoMean"`. It is based on a Binomial distribution, whose parameter is the geometric mean of the probabilities of success:
$$\hat{p} = \sqrt[n]{p_1 \cdot ... \cdot p_n}$$

```{r gma1}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
prod(rep(pp, wt))^(1/sum(wt))

dpbinom(NULL, pp, wt, "GeoMean")
ppbinom(NULL, pp, wt, "GeoMean")
```

It is known that the geometric mean of the probabilities of success is always smaller than their arithmetic mean. Thus, we get a stochastically *smaller* binomial distribution. A comparison with exact computation shows that the approximation quality of the GMBA-A procedure increases when the probabilities of success are closer to each other:

```{r gma2}
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "GeoMean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMean") - dpbinom(NULL, pp))

# U(0.4, 0.6) random probabilities of success
pp <- runif(20, 0.4, 0.6)
dpbinom(NULL, pp, method = "GeoMean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMean") - dpbinom(NULL, pp))

# U(0.49, 0.51) random probabilities of success
pp <- runif(20, 0.49, 0.51)
dpbinom(NULL, pp, method = "GeoMean")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMean") - dpbinom(NULL, pp))
```

### Geometric Mean Binomial Approximation - Variant B

The *Geometric Mean Binomial Approximation (Variant B)* (GMBA-B) approach is requested with `method = "GeoMeanCounter"`. It is based on a Binomial distribution, whose parameter is 1 minus the geometric mean of the probabilities of **failure**:
$$\hat{p} = 1 - \sqrt[n]{(1 - p_1) \cdot ... \cdot (1 - p_n)}$$

```{r gmb1}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
1 - prod(1 - rep(pp, wt))^(1/sum(wt))

dpbinom(NULL, pp, wt, "GeoMeanCounter")
ppbinom(NULL, pp, wt, "GeoMeanCounter")
```

It is known that the geometric mean of the probabilities of **failure** is always smaller than their arithmetic mean. As a result, 1 minus the geometric mean is larger than 1 minus the arithmetic mean. Thus, we get a stochastically *larger* binomial distribution. A comparison with exact computation shows that the approximation quality of the GMBA-B procedure again increases when the probabilities of success are closer to each other:

```{r gmb2}
set.seed(1)

# U(0, 1) random probabilities of success
pp <- runif(20)
dpbinom(NULL, pp, method = "GeoMeanCounter")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMeanCounter") - dpbinom(NULL, pp))

# U(0.4, 0.6) random probabilities of success
pp <- runif(20, 0.4, 0.6)
dpbinom(NULL, pp, method = "GeoMeanCounter")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMeanCounter") - dpbinom(NULL, pp))

# U(0.49, 0.51) random probabilities of success
pp <- runif(20, 0.49, 0.51)
dpbinom(NULL, pp, method = "GeoMeanCounter")
dpbinom(NULL, pp)
summary(dpbinom(NULL, pp, method = "GeoMeanCounter") - dpbinom(NULL, pp))
```

### Normal Approximation

The *Normal Approximation* (NA) approach is requested with `method = "Normal"`. It is based on a Normal distribution, whose parameters are derived from the theoretical mean and variance of the input probabilities of success.

```{r na1-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Normal")
ppbinom(NULL, pp, wt, "Normal")
```

A comparison with exact computation shows that the approximation quality of the NA procedure increases with larger numbers of probabilities of success:

```{r na2-ord}
set.seed(1)

# 10 random probabilities of success
pp <- runif(10)
dpn <- dpbinom(NULL, pp, method = "Normal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
dpn <- dpbinom(NULL, pp, method = "Normal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100000 random probabilities of success
pp <- runif(100000)
dpn <- dpbinom(NULL, pp, method = "Normal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])
```

### Refined Normal Approximation

The *Refined Normal Approximation* (RNA) approach is requested with `method = "RefinedNormal"`. It is based on a Normal distribution, whose parameters are derived from the theoretical mean, variance and skewness of the input probabilities of success.

```{r rna1-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "RefinedNormal")
ppbinom(NULL, pp, wt, "RefinedNormal")
```

A comparison with exact computation shows that the approximation quality of the RNA procedure increases with larger numbers of probabilities of success:

```{r rna2-ord}
set.seed(1)

# 10 random probabilities of success
pp <- runif(10)
dpn <- dpbinom(NULL, pp, method = "RefinedNormal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
dpn <- dpbinom(NULL, pp, method = "RefinedNormal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100000 random probabilities of success
pp <- runif(100000)
dpn <- dpbinom(NULL, pp, method = "RefinedNormal")
dpd <- dpbinom(NULL, pp)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])
```

### Processing Speed Comparisons

To assess the performance of the approximation procedures, we use the `microbenchmark` package. Each algorithm has to calculate the PMF repeatedly based on random probability vectors. The run times are then summarized in a table that presents, among other statistics, their minima, maxima and means. The following results were recorded on an AMD Ryzen 9 5900X with 64 GiB of RAM and Windows 10 Education (22H2).

```{r benchmark-ord}
library(microbenchmark)
set.seed(1)

f1 <- function() dpbinom(NULL, runif(4000), method = "Normal")
f2 <- function() dpbinom(NULL, runif(4000), method = "Poisson")
f3 <- function() dpbinom(NULL, runif(4000), method = "RefinedNormal")
f4 <- function() dpbinom(NULL, runif(4000), method = "Mean")
f5 <- function() dpbinom(NULL, runif(4000), method = "GeoMean")
f6 <- function() dpbinom(NULL, runif(4000), method = "GeoMeanCounter")
f7 <- function() dpbinom(NULL, runif(4000), method = "DivideFFT")

microbenchmark(f1(), f2(), f3(), f4(), f5(), f6(), f7(), times = 51)
```

Clearly, the NA procedure is the fastest, followed by the PA and RNA methods. The next fastest algorithms are AMBA, GMBA-A and GMBA-B. They exhibit almost equal mean execution speed, with the AMBA algorithm being slightly faster. All of the approximation procedures outperform the fastest exact approach, DC-FFT, by far.


## Generalized Poisson Binomial Distribution

### Generalized Normal Approximation

The *Generalized Normal Approximation* (G-NA) approach is requested with `method = "Normal"`. It is based on a Normal distribution, whose parameters are derived from the theoretical mean and variance of the input probabilities of success (see [Introduction](intro.html).

```{r na1-gen}
set.seed(2)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)

dgpbinom(NULL, pp, va, vb, wt, "Normal")
pgpbinom(NULL, pp, va, vb, wt, "Normal")
```

A comparison with exact computation shows that the approximation quality of the NA procedure increases with larger numbers of probabilities of success:

```{r na2-gen}
set.seed(2)

# 10 random probabilities of success
pp <- runif(10)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "Normal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100 random probabilities of success
pp <- runif(100)
va <- sample(0:100, 100, TRUE)
vb <- sample(0:100, 100, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "Normal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
va <- sample(0:1000, 1000, TRUE)
vb <- sample(0:1000, 1000, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "Normal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])
```

### Generalized Refined Normal Approximation

The *Generalized Refined Normal Approximation* (G-RNA) approach is requested with `method = "RefinedNormal"`. It is based on a Normal distribution, whose parameters are derived from the theoretical mean, variance and skewness of the input probabilities of success.

```{r rna1-gen}
set.seed(2)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)
dgpbinom(NULL, pp, va, vb, wt, "RefinedNormal")
pgpbinom(NULL, pp, va, vb, wt, "RefinedNormal")
```

A comparison with exact computation shows that the approximation quality of the RNA procedure increases with larger numbers of probabilities of success:

```{r rna2-gen}
set.seed(2)

# 10 random probabilities of success
pp <- runif(10)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 100 random probabilities of success
pp <- runif(100)
va <- sample(0:100, 100, TRUE)
vb <- sample(0:100, 100, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])

# 1000 random probabilities of success
pp <- runif(1000)
va <- sample(0:1000, 1000, TRUE)
vb <- sample(0:1000, 1000, TRUE)
dpn <- dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
dpd <- dgpbinom(NULL, pp, va, vb)
idx <- which(dpn != 0 & dpd != 0)
summary((dpn - dpd)[idx])
```

### Processing Speed Comparisons

To assess the performance of the approximation procedures, we use the `microbenchmark` package. Each algorithm has to calculate the PMF repeatedly based on random probability vectors. The run times are then summarized in a table that presents, among other statistics, their minima, maxima and means. The following results were recorded on an AMD Ryzen 9 5900X with 64 GiB of RAM and Windows 10 Education (22H2).

```{r benchmark-gen}
library(microbenchmark)
n <- 1500
set.seed(2)
va <- sample(1:50, n, TRUE)
vb <- sample(1:50, n, TRUE)

f1 <- function() dgpbinom(NULL, runif(n), va, vb, method = "Normal")
f2 <- function() dgpbinom(NULL, runif(n), va, vb, method = "RefinedNormal")
f3 <- function() dgpbinom(NULL, runif(n), va, vb, method = "DivideFFT")

microbenchmark(f1(), f2(), f3(), times = 51)
```

Clearly, the G-NA procedure is the fastest, followed by the G-RNA method. Both are hugely faster than G-DC-FFT.

---
title: "Exact Procedures"
output: 
  rmarkdown::html_vignette:
    toc: yes
vignette: >
  %\VignetteIndexEntry{Exact Procedures}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo = FALSE}
library(PoissonBinomial)
```

## Ordinary Poisson Binomial Distribution

### Direct Convolution

The *Direct Convolution* (DC) approach is requested with `method = "Convolve"`.

```{r directconv-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Convolve")
ppbinom(NULL, pp, wt, "Convolve")
```

### Divide & Conquer FFT Tree Convolution

The *Divide & Conquer FFT Tree Convolution* (DC-FFT) approach is requested with `method = "DivideFFT"`.

```{r divide1-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "DivideFFT")
ppbinom(NULL, pp, wt, "DivideFFT")
```

By design, as proposed by [Biscarri, Zhao & Brunner (2018)](http://dx.doi.org/10.1016/j.csda.2018.01.007), its results are identical to the DC procedure, if $n \leq 750$. Thus, differences can be observed for larger $n > 750$:

```{r divide2-ord}
set.seed(1)
pp1 <- runif(751)
pp2 <- pp1[1:750]

sum(abs(dpbinom(NULL, pp2, method = "DivideFFT") - dpbinom(NULL, pp2, method = "Convolve")))
sum(abs(dpbinom(NULL, pp1, method = "DivideFFT") - dpbinom(NULL, pp1, method = "Convolve")))
```

The reason is that the DC-FFT method splits the input `probs` vector into as equally sized parts as possible and computes their distributions separately with the DC approach. The results of the portions are then convoluted by means of the Fast Fourier Transformation. As proposed by [Biscarri, Zhao & Brunner (2018)](http://dx.doi.org/10.1016/j.csda.2018.01.007), no splitting is done for $n \leq 750$. In addition, the DC-FFT procedure does not produce probabilities $\leq 5.55e\text{-}17$, i.e. smaller values are rounded off to 0, if $n > 750$, whereas the smallest possible result of the DC algorithm is $\sim 1e\text{-}323$. This is most likely caused by the used FFTW3 library.

```{r divide3-ord}
set.seed(1)
pp1 <- runif(751)

d1 <- dpbinom(NULL, pp1, method = "DivideFFT")
d2 <- dpbinom(NULL, pp1, method = "Convolve")

min(d1[d1 > 0])
min(d2[d2 > 0])
```

### Discrete Fourier Transformation of the Characteristic Function

The *Discrete Fourier Transformation of the Characteristic Function* (DFT-CF) approach is requested with `method = "Characteristic"`.

```{r dftcf-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Characteristic")
ppbinom(NULL, pp, wt, "Characteristic")
```

As can be seen, the DFT-CF procedure does not produce probabilities $\leq 2.22e\text{-}16$, i.e. smaller values are rounded off to 0, most likely due to the used FFTW3 library.

### Recursive Formula

The *Recursive Formula* (RF) approach is requested with `method = "Recursive"`.

```{r rf1-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Recursive")
ppbinom(NULL, pp, wt, "Recursive")
```

Obviously, the RF procedure does produce probabilities $\leq 5.55e\text{-}17$, because it does not rely on the FFTW3 library. Furthermore, it yields the same results as the DC method.

```{r rf2-ord}
set.seed(1)
pp <- runif(1000)
wt <- sample(1:10, 1000, TRUE)

sum(abs(dpbinom(NULL, pp, wt, "Convolve") - dpbinom(NULL, pp, wt, "Recursive")))
```

### Processing Speed Comparisons

To assess the performance of the exact procedures, we use the `microbenchmark` package. Each algorithm has to calculate the PMF repeatedly based on random probability vectors. The run times are then summarized in a table that presents, among other statistics, their minima, maxima and means. The following results were recorded on an AMD Ryzen 9 5900X with 64 GiB of RAM and Manjaro KDE Linux (26.0.2).

```{r benchmark-ord}
library(microbenchmark)
set.seed(1)

f1 <- function() dpbinom(NULL, runif(6000), method = "DivideFFT")
f2 <- function() dpbinom(NULL, runif(6000), method = "Convolve")
f3 <- function() dpbinom(NULL, runif(6000), method = "Recursive")
f4 <- function() dpbinom(NULL, runif(6000), method = "Characteristic")

microbenchmark(f1(), f2(), f3(), f4(), times = 51)
```

Clearly, the DC-FFT procedure is the fastest, followed by DC, RF and DFT-CF methods.


## Generalized Poisson Binomial Distribution

### Generalized Direct Convolution

The *Generalized Direct Convolution* (G-DC) approach is requested with `method = "Convolve"`.

```{r directconv-gen}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)

dgpbinom(NULL, pp, va, vb, wt, "Convolve")
pgpbinom(NULL, pp, va, vb, wt, "Convolve")
```

### Generalized Divide & Conquer FFT Tree Convolution

The *Generalized Divide & Conquer FFT Tree Convolution* (G-DC-FFT) approach is requested with `method = "DivideFFT"`.

```{r divide1-gen}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)

dgpbinom(NULL, pp, va, vb, wt, "DivideFFT")
pgpbinom(NULL, pp, va, vb, wt, "DivideFFT")
```

By design, similar to the ordinary DC-FFT algorithm by [Biscarri, Zhao & Brunner (2018)](http://dx.doi.org/10.1016/j.csda.2018.01.007), its results are identical to the G-DC procedure, if $n$ and the number of possible observed values is small. Thus, differences can be observed for larger numbers:

```{r divide2-gen}
set.seed(1)
pp1 <- runif(250)
va1 <- sample(0:50, 250, TRUE)
vb1 <- sample(0:50, 250, TRUE)
pp2 <- pp1[1:248]
va2 <- va1[1:248]
vb2 <- vb1[1:248]

sum(abs(dgpbinom(NULL, pp1, va1, vb1, method = "DivideFFT")
        - dgpbinom(NULL, pp1, va1, vb1, method = "Convolve")))

sum(abs(dgpbinom(NULL, pp2, va2, vb2, method = "DivideFFT")
        - dgpbinom(NULL, pp2, va2, vb2, method = "Convolve")))
```

The reason is that the G-DC-FFT method splits the input `probs`, `val_p` and `val_q` vectors into parts such that the numbers of possible observations of all parts are as equally sized as possible. Their distributions are then computed separately with the G-DC approach. The results of the portions are then convoluted by means of the Fast Fourier Transformation. For small $n$ and small distribution sizes, no splitting is needed. In addition, the G-DC-FFT procedure, just like the DC-FFT method, does not produce probabilities $\leq 5.55e\text{-}17$, i.e. smaller values are rounded off to $0$, if the total number of possible observations is smaller than $750$, whereas the smallest possible result of the DC algorithm is $\sim 1e\text{-}323$. This is most likely caused by the used FFTW3 library.

```{r divide3-gen}
d1 <- dgpbinom(NULL, pp1, va1, vb1, method = "DivideFFT")
d2 <- dgpbinom(NULL, pp1, va1, vb1, method = "Convolve")

min(d1[d1 > 0])
min(d2[d2 > 0])
```

### Generalized Discrete Fourier Transformation of the Characteristic Function

The *Generalized Discrete Fourier Transformation of the Characteristic Function* (G-DFT-CF) approach is requested with `method = "Characteristic"`.

```{r dftcf-gen}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)

dgpbinom(NULL, pp, va, vb, wt, "Characteristic")
pgpbinom(NULL, pp, va, vb, wt, "Characteristic")
```

As can be seen, the G-DFT-CF procedure does not produce probabilities $\leq 2.2e\text{-}16$, i.e. smaller values are rounded off to 0, most likely due to the used FFTW3 library.

### Processing Speed Comparisons

To assess the performance of the exact procedures, we use the `microbenchmark` package. Each algorithm has to calculate the PMF repeatedly based on random probability and value vectors. The run times are then summarized in a table that presents, among other statistics, their minima, maxima and means. The following results were recorded on an AMD Ryzen 9 5900X with 64 GiB of RAM and Manjaro KDE Linux (26.0.2).

```{r benchmark-gen}
library(microbenchmark)
n <- 2500
set.seed(1)
va <- sample(1:50, n, TRUE)
vb <- sample(1:50, n, TRUE)

f1 <- function() dgpbinom(NULL, runif(n), va, vb, method = "DivideFFT")
f2 <- function() dgpbinom(NULL, runif(n), va, vb, method = "Convolve")
f3 <- function() dgpbinom(NULL, runif(n), va, vb, method = "Characteristic")

microbenchmark(f1(), f2(), f3(), times = 51)
```

Clearly, the G-DC-FFT procedure is the fastest one. It outperforms both the G-DC and G-DFT-CF approaches. The latter one needs a lot more time than the others. Generally, the computational speed advantage of the G-DC-FFT procedure increases with larger $n$ (and $m$).

---
title: "Usage with Rcpp"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Usage with Rcpp}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

Each procedure's probability mass function (PMF) and cumulative distribution function (CDF) was implemented in *C++* using the `Rcpp` package. By means of `Rcpp::interface`, these functions are exported to both the package's *R* namespace and *C++* headers. That way, the following functions can then be used by other packages that use `Rcpp`:

```
/***   Ordinary Poisson Binomial Distribution   ***/


/***   Exact Procedures   ***/

// Direct Convolution (DC)

// PMF
NumericVector dpb_conv(const IntegerVector obs,
                       const NumericVector probs);
                       
// CDF
NumericVector ppb_conv(const IntegerVector obs,
                       const NumericVector probs,
                       const bool lower_tail);


// Divide & Conquer FFT Tree Convolution (DC-FFT)

// PMF
NumericVector dpb_dc(const IntegerVector obs,
                     const NumericVector probs);
                     
// CDF
NumericVector ppb_dc(const IntegerVector obs,
                     const NumericVector probs,
                     const bool lower_tail);


// Discrete Fourier Transformation of the Characteristic Function (DFT-CF)

// PMF
NumericVector dpb_dftcf(const IntegerVector obs,
                        const NumericVector probs);
                        
// CDF
NumericVector ppb_dftcf(const IntegerVector obs, const NumericVector probs,
                        const bool lower_tail);
                        

// Recursive Formula (RF)

// PMF
NumericVector dpb_rf(const IntegerVector obs,
                     const NumericVector probs);

// CDF
NumericVector ppb_rf(const IntegerVector obs,
                     const NumericVector probs,
                     const bool lower_tail);



/***   Approximations   ***/


// Arithmetic Mean Binomial Approximation (AMBA)

// PMF
NumericVector dpb_mean(const IntegerVector obs,
                       const NumericVector probs);

// CDF
NumericVector ppb_mean(const IntegerVector obs,
                       const NumericVector probs,
                       const bool lower_tail);


// Geometric Mean Binomial Approximations (GMBA)

// PMF
NumericVector dpb_gmba(const IntegerVector obs, 
                       const NumericVector const probs,
                       const bool anti);
                       
// CDF
NumericVector ppb_gmba(const IntegerVector obs,
                       const NumericVector probs,
                       const bool anti,
                       const bool lower_tail);


// Poisson Approximation (PA)

// PMF
NumericVector dpb_pa(const IntegerVector obs,
                     const NumericVector probs);
                     
// CDF
NumericVector ppb_pa(const IntegerVector obs,
                     const NumericVector probs,
                     const bool lower_tail);
                     

// Normal Approximations (NA, RNA)

// PMF
NumericVector dpb_na(const IntegerVector obs,
                     const NumericVector probs,
                     const bool refined);
                     
// CDF
NumericVector ppb_na(const IntegerVector obs,
                     const NumericVector probs,
                     const bool refined,
                     const bool lower_tail);
                     



/***   Generalized Poisson Binomial Distribution   ***/


/***   Exact Procedures   ***/


// Generalized Direct Convolution (G-DC)

// PMF
NumericVector dgpb_conv(const IntegerVector obs,
                        const NumericVector probs,
                        const NumericVector val_p,
                        const NumericVector val_q);
                        
// CDF
NumericVector pgpb_conv(const IntegerVector obs,
                        const NumericVector probs,
                        const NumericVector val_p,
                        const NumericVector val_q,
                        const bool lower_tail);
                        

// Generalized Discrete Fourier Transformation of the Characteristic Function (G-DFT-CF)

// PMF
NumericVector dgpb_dftcf(const IntegerVector obs,
                         const NumericVector probs,
                         const NumericVector val_p,
                         const NumericVector val_q);
                         
// CDF
NumericVector pgpb_dftcf(const IntegerVector obs,
                         const NumericVector probs,
                         const NumericVector val_p,
                         const NumericVector val_q,
                         const bool lower_tail);
                       
                       
                       
/***   Approximations   ***/


// Generalized Normal Approximations (G-NA, G-RNA)

// PMF
NumericVector dgpb_na(const IntegerVector obs,
                      const NumericVector probs,
                      const NumericVector val_p,
                      const NumericVector val_q,
                      const bool refined,
                      const bool lower_tail);
                      
// CDF
NumericVector pgpb_na(const IntegerVector obs,
                      const NumericVector probs,
                      const NumericVector val_p,
                      const NumericVector val_q,
                      const bool refined,
                      const bool lower_tail);
```


## Making the functions usable

There are only a few simple steps to follow:

1. Add the `Rcpp` and `PoissonBinomial` packages to the `Imports` and `LinkingTo` fields of the `DESCRIPTION` file.
2. Add `#include <PoissonBinomial.h>` to source (`.cpp`) and/or header (`.h`, `.hpp`) files in which these functions are to be used.
3. Optional: Add `using namespace PoissonBinomial;`. Without it, the use of functions of this package must be fully qualified with `PoissonBinomial::`, e.g. `PoissonBinomial::dpb_dc` instead of `dpb_dc`


## Important Remarks

For better performance, the PMFs and CDFs do not check any of their parameters for plausibility! This must be done by the user by means of *R* or *C/C++* functions. It must be made sure that

* the observations in the `obs` vectors are valid, 
* the probabilities in the `probs` vector are in $(0, 1)$ and
* for `dpb_gmba`, `ppb_gmba`, `dpb_na`, `ppb_na`, `dgpb_na` and `pgpb_na`: the probabilities in the `probs` vector **must not** contain zeros or ones.

Furthermore, the CDFs only compute non-logarithmic probabilities. If logarithms are needed, they must be computed "manually".

---
title: "Exact Procedures"
output: 
  rmarkdown::html_vignette:
    toc: yes
vignette: >
  %\VignetteIndexEntry{Exact Procedures}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup, echo = FALSE}
library(PoissonBinomial)
```

## Ordinary Poisson Binomial Distribution

### Direct Convolution

The *Direct Convolution* (DC) approach is requested with `method = "Convolve"`.

```{r directconv-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Convolve")
ppbinom(NULL, pp, wt, "Convolve")
```

### Divide & Conquer FFT Tree Convolution

The *Divide & Conquer FFT Tree Convolution* (DC-FFT) approach is requested with `method = "DivideFFT"`.

```{r divide1-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "DivideFFT")
ppbinom(NULL, pp, wt, "DivideFFT")
```

By design, as proposed by [Biscarri, Zhao & Brunner (2018)](http://dx.doi.org/10.1016/j.csda.2018.01.007), its results are identical to the DC procedure, if $n \leq 750$. Thus, differences can be observed for larger $n > 750$:

```{r divide2-ord}
set.seed(1)
pp1 <- runif(751)
pp2 <- pp1[1:750]

sum(abs(dpbinom(NULL, pp2, method = "DivideFFT") - dpbinom(NULL, pp2, method = "Convolve")))
sum(abs(dpbinom(NULL, pp1, method = "DivideFFT") - dpbinom(NULL, pp1, method = "Convolve")))
```

The reason is that the DC-FFT method splits the input `probs` vector into as equally sized parts as possible and computes their distributions separately with the DC approach. The results of the portions are then convoluted by means of the Fast Fourier Transformation. As proposed by [Biscarri, Zhao & Brunner (2018)](http://dx.doi.org/10.1016/j.csda.2018.01.007), no splitting is done for $n \leq 750$. In addition, the DC-FFT procedure does not produce probabilities $\leq 5.55e\text{-}17$, i.e. smaller values are rounded off to 0, if $n > 750$, whereas the smallest possible result of the DC algorithm is $\sim 1e\text{-}323$. This is most likely caused by the used FFTW3 library.

```{r divide3-ord}
set.seed(1)
pp1 <- runif(751)

d1 <- dpbinom(NULL, pp1, method = "DivideFFT")
d2 <- dpbinom(NULL, pp1, method = "Convolve")

min(d1[d1 > 0])
min(d2[d2 > 0])
```

### Discrete Fourier Transformation of the Characteristic Function

The *Discrete Fourier Transformation of the Characteristic Function* (DFT-CF) approach is requested with `method = "Characteristic"`.

```{r dftcf-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Characteristic")
ppbinom(NULL, pp, wt, "Characteristic")
```

As can be seen, the DFT-CF procedure does not produce probabilities $\leq 2.22e\text{-}16$, i.e. smaller values are rounded off to 0, most likely due to the used FFTW3 library.

### Recursive Formula

The *Recursive Formula* (RF) approach is requested with `method = "Recursive"`.

```{r rf1-ord}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)

dpbinom(NULL, pp, wt, "Recursive")
ppbinom(NULL, pp, wt, "Recursive")
```

Obviously, the RF procedure does produce probabilities $\leq 5.55e\text{-}17$, because it does not rely on the FFTW3 library. Furthermore, it yields the same results as the DC method.

```{r rf2-ord}
set.seed(1)
pp <- runif(1000)
wt <- sample(1:10, 1000, TRUE)

sum(abs(dpbinom(NULL, pp, wt, "Convolve") - dpbinom(NULL, pp, wt, "Recursive")))
```

### Processing Speed Comparisons

To assess the performance of the exact procedures, we use the `microbenchmark` package. Each algorithm has to calculate the PMF repeatedly based on random probability vectors. The run times are then summarized in a table that presents, among other statistics, their minima, maxima and means. The following results were recorded on an AMD Ryzen 9 5900X with 64 GiB of RAM and Windows 10 Education (22H2).

```{r benchmark-ord}
library(microbenchmark)
set.seed(1)

f1 <- function() dpbinom(NULL, runif(6000), method = "DivideFFT")
f2 <- function() dpbinom(NULL, runif(6000), method = "Convolve")
f3 <- function() dpbinom(NULL, runif(6000), method = "Recursive")
f4 <- function() dpbinom(NULL, runif(6000), method = "Characteristic")

microbenchmark(f1(), f2(), f3(), f4(), times = 51)
```

Clearly, the DC-FFT procedure is the fastest, followed by DC, RF and DFT-CF methods.


## Generalized Poisson Binomial Distribution

### Generalized Direct Convolution

The *Generalized Direct Convolution* (G-DC) approach is requested with `method = "Convolve"`.

```{r directconv-gen}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)

dgpbinom(NULL, pp, va, vb, wt, "Convolve")
pgpbinom(NULL, pp, va, vb, wt, "Convolve")
```

### Generalized Divide & Conquer FFT Tree Convolution

The *Generalized Divide & Conquer FFT Tree Convolution* (G-DC-FFT) approach is requested with `method = "DivideFFT"`.

```{r divide1-gen}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)

dgpbinom(NULL, pp, va, vb, wt, "DivideFFT")
pgpbinom(NULL, pp, va, vb, wt, "DivideFFT")
```

By design, similar to the ordinary DC-FFT algorithm by [Biscarri, Zhao & Brunner (2018)](http://dx.doi.org/10.1016/j.csda.2018.01.007), its results are identical to the G-DC procedure, if $n$ and the number of possible observed values is small. Thus, differences can be observed for larger numbers:

```{r divide2-gen}
set.seed(1)
pp1 <- runif(250)
va1 <- sample(0:50, 250, TRUE)
vb1 <- sample(0:50, 250, TRUE)
pp2 <- pp1[1:248]
va2 <- va1[1:248]
vb2 <- vb1[1:248]

sum(abs(dgpbinom(NULL, pp1, va1, vb1, method = "DivideFFT")
        - dgpbinom(NULL, pp1, va1, vb1, method = "Convolve")))

sum(abs(dgpbinom(NULL, pp2, va2, vb2, method = "DivideFFT")
        - dgpbinom(NULL, pp2, va2, vb2, method = "Convolve")))
```

The reason is that the G-DC-FFT method splits the input `probs`, `val_p` and `val_q` vectors into parts such that the numbers of possible observations of all parts are as equally sized as possible. Their distributions are then computed separately with the G-DC approach. The results of the portions are then convoluted by means of the Fast Fourier Transformation. For small $n$ and small distribution sizes, no splitting is needed. In addition, the G-DC-FFT procedure, just like the DC-FFT method, does not produce probabilities $\leq 5.55e\text{-}17$, i.e. smaller values are rounded off to $0$, if the total number of possible observations is smaller than $750$, whereas the smallest possible result of the DC algorithm is $\sim 1e\text{-}323$. This is most likely caused by the used FFTW3 library.

```{r divide3-gen}
d1 <- dgpbinom(NULL, pp1, va1, vb1, method = "DivideFFT")
d2 <- dgpbinom(NULL, pp1, va1, vb1, method = "Convolve")

min(d1[d1 > 0])
min(d2[d2 > 0])
```

### Generalized Discrete Fourier Transformation of the Characteristic Function

The *Generalized Discrete Fourier Transformation of the Characteristic Function* (G-DFT-CF) approach is requested with `method = "Characteristic"`.

```{r dftcf-gen}
set.seed(1)
pp <- runif(10)
wt <- sample(1:10, 10, TRUE)
va <- sample(0:10, 10, TRUE)
vb <- sample(0:10, 10, TRUE)

dgpbinom(NULL, pp, va, vb, wt, "Characteristic")
pgpbinom(NULL, pp, va, vb, wt, "Characteristic")
```

As can be seen, the G-DFT-CF procedure does not produce probabilities $\leq 2.2e\text{-}16$, i.e. smaller values are rounded off to 0, most likely due to the used FFTW3 library.

### Processing Speed Comparisons

To assess the performance of the exact procedures, we use the `microbenchmark` package. Each algorithm has to calculate the PMF repeatedly based on random probability and value vectors. The run times are then summarized in a table that presents, among other statistics, their minima, maxima and means. The following results were recorded on an AMD Ryzen 9 5900X with 64 GiB of RAM and Windows 10 Education (22H2).

```{r benchmark-gen}
library(microbenchmark)
n <- 2500
set.seed(1)
va <- sample(1:50, n, TRUE)
vb <- sample(1:50, n, TRUE)

f1 <- function() dgpbinom(NULL, runif(n), va, vb, method = "DivideFFT")
f2 <- function() dgpbinom(NULL, runif(n), va, vb, method = "Convolve")
f3 <- function() dgpbinom(NULL, runif(n), va, vb, method = "Characteristic")

microbenchmark(f1(), f2(), f3(), times = 51)
```

Clearly, the G-DC-FFT procedure is the fastest one. It outperforms both the G-DC and G-DFT-CF approaches. The latter one needs a lot more time than the others. Generally, the computational speed advantage of the G-DC-FFT procedure increases with larger $n$ (and $m$).

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gpbinom.R
\name{GenPoissonBinomial-Distribution}
\alias{GenPoissonBinomial-Distribution}
\alias{dgpbinom}
\alias{pgpbinom}
\alias{qgpbinom}
\alias{rgpbinom}
\title{The Generalized Poisson Binomial Distribution}
\usage{
dgpbinom(x, probs, val_p, val_q, wts = NULL, method = "DivideFFT", log = FALSE)

pgpbinom(
  x,
  probs,
  val_p,
  val_q,
  wts = NULL,
  method = "DivideFFT",
  lower.tail = TRUE,
  log.p = FALSE
)

qgpbinom(
  p,
  probs,
  val_p,
  val_q,
  wts = NULL,
  method = "DivideFFT",
  lower.tail = TRUE,
  log.p = FALSE
)

rgpbinom(
  n,
  probs,
  val_p,
  val_q,
  wts = NULL,
  method = "DivideFFT",
  generator = "Sample"
)
}
\arguments{
\item{x}{Either a vector of observed sums or NULL. If NULL,
probabilities of all possible observations are
returned.}

\item{probs}{Vector of probabilities of success of each Bernoulli
trial.}

\item{val_p}{Vector of values that each trial produces with probability
in \code{probs}.}

\item{val_q}{Vector of values that each trial produces with probability
in \code{1 - probs}.}

\item{wts}{Vector of non-negative integer weights for the input
probabilities.}

\item{method}{Character string that specifies the method of computation
and must be one of \code{"DivideFFT"}, \code{"Convolve"}, 
\code{"Characteristic"}, \code{"Normal"} or
\code{"RefinedNormal"} (abbreviations are allowed).}

\item{log, log.p}{Logical value indicating if results are given as
logarithms.}

\item{lower.tail}{Logical value indicating if results are \eqn{P[X \leq x]}
(if \code{TRUE}; default) or \eqn{P[X > x]} (if 
\code{FALSE}).}

\item{p}{Vector of probabilities for computation of quantiles.}

\item{n}{Number of observations. If \code{length(n) > 1}, the
length is taken to be the number required.}

\item{generator}{Character string that specifies the random number
generator and must either be \code{"Sample"} or
\code{"Bernoulli"} (abbreviations are allowed).}
}
\value{
\code{dgpbinom} gives the density, \code{pgpbinom} computes the distribution
function, \code{qgpbinom} gives the quantile function and \code{rgpbinom}
generates random deviates.

For \code{rgpbinom}, the length of the result is determined by \code{n}, and
is the lengths of the numerical arguments for the other functions.
}
\description{
Density, distribution function, quantile function and random generation for
the generalized Poisson binomial distribution with probability vector
\code{probs}.
}
\details{
See the references for computational details. The \emph{Divide and Conquer}
(\code{"DivideFFT"}) and \emph{Direct Convolution} (\code{"Convolve"})
algorithms are derived and described in Biscarri, Zhao & Brunner (2018). They
have been modified for use with the generalized Poisson binomial
distribution. The
\emph{Discrete Fourier Transformation of the Characteristic Function}
(\code{"Characteristic"}) is derived in Zhang, Hong & Balakrishnan (2018),
the \emph{Normal Approach} (\code{"Normal"}) and the
\emph{Refined Normal Approach} (\code{"RefinedNormal"}) are described in Hong
(2013). They were slightly adapted for the generalized Poisson binomial
distribution.

In some special cases regarding the values of \code{probs}, the \code{method}
parameter is ignored (see Introduction vignette).

Random numbers can be generated in two ways. The \code{"Sample"} method
uses \code{R}'s \code{sample} function to draw random values according to
their probabilities that are calculated by \code{dgpbinom}. The
\code{"Bernoulli"} procedure ignores the \code{method} parameter and
simulates Bernoulli-distributed random numbers according to the probabilities
in \code{probs} and sums them up. It is a bit slower than the \code{"Sample"}
generator, but may yield better results, as it allows to obtain observations
that cannot be generated by the \code{"Sample"} procedure, because
\code{dgpbinom} may compute 0-probabilities, due to rounding, if the length
of \code{probs} is large and/or its values contain a lot of very small
values.
}
\section{References}{

Hong, Y. (2018). On computing the distribution function for the Poisson
   binomial distribution. \emph{Computational Statistics & Data Analysis},
   \strong{59}, pp. 41-51. doi:
   \href{https://doi.org/10.1016/j.csda.2012.10.006}{
   10.1016/j.csda.2012.10.006}

Biscarri, W., Zhao, S. D. and Brunner, R. J. (2018) A simple and fast method
   for computing the Poisson binomial distribution.
   \emph{Computational Statistics and Data Analysis}, \strong{31}, pp.
   216–222. doi: \href{https://doi.org/10.1016/j.csda.2018.01.007}{
   10.1016/j.csda.2018.01.007}
   
Zhang, M., Hong, Y. and Balakrishnan, N. (2018). The generalized 
   Poisson-binomial distribution and the computation of its distribution
   function. \emph{Journal of Statistical Computational and Simulation},
   \strong{88}(8), pp. 1515-1527. doi:
   \href{https://doi.org/10.1080/00949655.2018.1440294}{
   10.1080/00949655.2018.1440294}
}

\examples{
set.seed(1)
pp <- c(1, 0, runif(10), 1, 0, 1)
qq <- seq(0, 1, 0.01)
va <- rep(5, length(pp))
vb <- 1:length(pp)

dgpbinom(NULL, pp, va, vb, method = "DivideFFT")
pgpbinom(75:100, pp, va, vb, method = "DivideFFT")
qgpbinom(qq, pp, va, vb, method = "DivideFFT")
rgpbinom(100, pp, va, vb, method = "DivideFFT")

dgpbinom(NULL, pp, va, vb, method = "Convolve")
pgpbinom(75:100, pp, va, vb, method = "Convolve")
qgpbinom(qq, pp, va, vb, method = "Convolve")
rgpbinom(100, pp, va, vb, method = "Convolve")

dgpbinom(NULL, pp, va, vb, method = "Characteristic")
pgpbinom(75:100, pp, va, vb, method = "Characteristic")
qgpbinom(qq, pp, va, vb, method = "Characteristic")
rgpbinom(100, pp, va, vb, method = "Characteristic")

dgpbinom(NULL, pp, va, vb, method = "Normal")
pgpbinom(75:100, pp, va, vb, method = "Normal")
qgpbinom(qq, pp, va, vb, method = "Normal")
rgpbinom(100, pp, va, vb, method = "Normal")

dgpbinom(NULL, pp, va, vb, method = "RefinedNormal")
pgpbinom(75:100, pp, va, vb, method = "RefinedNormal")
qgpbinom(qq, pp, va, vb, method = "RefinedNormal")
rgpbinom(100, pp, va, vb, method = "RefinedNormal")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pbinom.R
\name{PoissonBinomial-Distribution}
\alias{PoissonBinomial-Distribution}
\alias{dpbinom}
\alias{ppbinom}
\alias{qpbinom}
\alias{rpbinom}
\title{The Poisson Binomial Distribution}
\usage{
dpbinom(x, probs, wts = NULL, method = "DivideFFT", log = FALSE)

ppbinom(
  x,
  probs,
  wts = NULL,
  method = "DivideFFT",
  lower.tail = TRUE,
  log.p = FALSE
)

qpbinom(
  p,
  probs,
  wts = NULL,
  method = "DivideFFT",
  lower.tail = TRUE,
  log.p = FALSE
)

rpbinom(n, probs, wts = NULL, method = "DivideFFT", generator = "Sample")
}
\arguments{
\item{x}{Either a vector of observed numbers of successes or NULL.
If NULL, probabilities of all possible observations are
returned.}

\item{probs}{Vector of probabilities of success of each Bernoulli
trial.}

\item{wts}{Vector of non-negative integer weights for the input
probabilities.}

\item{method}{Character string that specifies the method of computation
and must be one of \code{"DivideFFT"}, \code{"Convolve"},
\code{"Characteristic"}, \code{"Recursive"},
\code{"Mean"}, \code{"GeoMean"}, \code{"GeoMeanCounter"},
\code{"Poisson"}, \code{"Normal"} or
\code{"RefinedNormal"} (abbreviations are allowed).}

\item{log, log.p}{Logical value indicating if results are given as
logarithms.}

\item{lower.tail}{Logical value indicating if results are \eqn{P[X \leq x]}
(if \code{TRUE}; default) or \eqn{P[X > x]} (if 
\code{FALSE}).}

\item{p}{Vector of probabilities for computation of quantiles.}

\item{n}{Number of observations. If \code{length(n) > 1}, the
length is taken to be the number required.}

\item{generator}{Character string that specifies the random number
generator and must either be \code{"Sample"} (default) or
\code{"Bernoulli"} (abbreviations are allowed). See
Details for more information.}
}
\value{
\code{dpbinom} gives the density, \code{ppbinom} computes the distribution
function, \code{qpbinom} gives the quantile function and \code{rpbinom}
generates random deviates.

For \code{rpbinom}, the length of the result is determined by \code{n}, and
is the lengths of the numerical arguments for the other functions.
}
\description{
Density, distribution function, quantile function and random generation for
the Poisson binomial distribution with probability vector \code{probs}.
}
\details{
See the references for computational details. The \emph{Divide and Conquer}
(\code{"DivideFFT"}) and \emph{Direct Convolution} (\code{"Convolve"})
algorithms are derived and described in Biscarri, Zhao & Brunner (2018). The
\emph{Discrete Fourier Transformation of the Characteristic Function}
(\code{"Characteristic"}), the \emph{Recursive Formula} (\code{"Recursive"}),
the \emph{Poisson Approximation} (\code{"Poisson"}), the
\emph{Normal Approach} (\code{"Normal"}) and the
\emph{Refined Normal Approach} (\code{"RefinedNormal"}) are described in Hong
(2013). The calculation of the \emph{Recursive Formula} was modified to
overcome the excessive memory requirements of Hong's implementation.

The \code{"Mean"} method is a naive binomial approach using the arithmetic
mean of the probabilities of success. Similarly, the \code{"GeoMean"} and
\code{"GeoMeanCounter"} procedures are binomial approximations, too, but
they form the geometric mean of the probabilities of success
(\code{"GeoMean"}) and their counter probabilities (\code{"GeoMeanCounter"}),
respectively.

In some special cases regarding the values of \code{probs}, the \code{method}
parameter is ignored (see Introduction vignette).

Random numbers can be generated in two ways. The \code{"Sample"} method
uses \code{R}'s \code{sample} function to draw random values according to
their probabilities that are calculated by \code{dgpbinom}. The
\code{"Bernoulli"} procedure ignores the \code{method} parameter and
simulates Bernoulli-distributed random numbers according to the probabilities
in \code{probs} and sums them up. It is a bit slower than the \code{"Sample"}
generator, but may yield better results, as it allows to obtain observations
that cannot be generated by the \code{"Sample"} procedure, because
\code{dgpbinom} may compute 0-probabilities, due to rounding, if the length
of \code{probs} is large and/or its values contain a lot of very small
values.
}
\section{References}{

Hong, Y. (2013). On computing the distribution function for the Poisson
   binomial distribution. \emph{Computational Statistics & Data Analysis},
   \strong{59}, pp. 41-51. \doi{10.1016/j.csda.2012.10.006}

Biscarri, W., Zhao, S. D. and Brunner, R. J. (2018) A simple and fast method
   for computing the Poisson binomial distribution.
   \emph{Computational Statistics and Data Analysis}, \strong{31}, pp.
   216–222. \doi{10.1016/j.csda.2018.01.007}
}

\examples{
set.seed(1)
pp <- c(0, 0, runif(995), 1, 1, 1)
qq <- seq(0, 1, 0.01)

dpbinom(NULL, pp, method = "DivideFFT")
ppbinom(450:550, pp, method = "DivideFFT")
qpbinom(qq, pp, method = "DivideFFT")
rpbinom(100, pp, method = "DivideFFT")

dpbinom(NULL, pp, method = "Convolve")
ppbinom(450:550, pp, method = "Convolve")
qpbinom(qq, pp, method = "Convolve")
rpbinom(100, pp, method = "Convolve")

dpbinom(NULL, pp, method = "Characteristic")
ppbinom(450:550, pp, method = "Characteristic")
qpbinom(qq, pp, method = "Characteristic")
rpbinom(100, pp, method = "Characteristic")

dpbinom(NULL, pp, method = "Recursive")
ppbinom(450:550, pp, method = "Recursive")
qpbinom(qq, pp, method = "Recursive")
rpbinom(100, pp, method = "Recursive")

dpbinom(NULL, pp, method = "Mean")
ppbinom(450:550, pp, method = "Mean")
qpbinom(qq, pp, method = "Mean")
rpbinom(100, pp, method = "Mean")

dpbinom(NULL, pp, method = "GeoMean")
ppbinom(450:550, pp, method = "GeoMean")
qpbinom(qq, pp, method = "GeoMean")
rpbinom(100, pp, method = "GeoMean")

dpbinom(NULL, pp, method = "GeoMeanCounter")
ppbinom(450:550, pp, method = "GeoMeanCounter")
qpbinom(qq, pp, method = "GeoMeanCounter")
rpbinom(100, pp, method = "GeoMeanCounter")

dpbinom(NULL, pp, method = "Poisson")
ppbinom(450:550, pp, method = "Poisson")
qpbinom(qq, pp, method = "Poisson")
rpbinom(100, pp, method = "Poisson")

dpbinom(NULL, pp, method = "Normal")
ppbinom(450:550, pp, method = "Normal")
qpbinom(qq, pp, method = "Normal")
rpbinom(100, pp, method = "Normal")

dpbinom(NULL, pp, method = "RefinedNormal")
ppbinom(450:550, pp, method = "RefinedNormal")
qpbinom(qq, pp, method = "RefinedNormal")
rpbinom(100, pp, method = "RefinedNormal")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/PoissonBinomial.R
\docType{package}
\name{PoissonBinomial-package}
\alias{PoissonBinomial-package}
\title{Efficient Exact and Approximate Implementations for Computing Ordinary and Generalized Poisson Binomial Distributions}
\description{
This package implements various algorithms for computing the probability mass
function, the cumulative distribution function, quantiles and random numbers
of both ordinary and generalized Poisson binomial distributions.
}
\section{References}{

Hong, Y. (2013). On computing the distribution function for the Poisson
   binomial distribution. \emph{Computational Statistics & Data Analysis},
   \strong{59}, pp. 41-51. \doi{10.1016/j.csda.2012.10.006}

Biscarri, W., Zhao, S. D. and Brunner, R. J. (2018) A simple and fast method
   for computing the Poisson binomial distribution.
   \emph{Computational Statistics and Data Analysis}, \strong{31}, pp.
   216–222. \doi{10.1016/j.csda.2018.01.007}
   
Zhang, M., Hong, Y. and Balakrishnan, N. (2018). The generalized 
   Poisson-binomial distribution and the computation of its distribution
   function. \emph{Journal of Statistical Computational and Simulation},
   \strong{88}(8), pp. 1515-1527. \doi{10.1080/00949655.2018.1440294}
}

\examples{
# Functions for ordinary Poisson binomial distributions
set.seed(1)
pp <- c(1, 0, runif(10), 1, 0, 1)
qq <- seq(0, 1, 0.01)

dpbinom(NULL, pp)
ppbinom(7:10, pp, method = "DivideFFT")
qpbinom(qq, pp, method = "Convolve")
rpbinom(10, pp, method = "RefinedNormal")

# Functions for generalized Poisson binomial distributions
va <- rep(5, length(pp))
vb <- 1:length(pp)

dgpbinom(NULL, pp, va, vb, method = "Convolve")
pgpbinom(80:100, pp, va, vb, method = "Convolve")
qgpbinom(qq, pp, va, vb, method = "Convolve")
rgpbinom(100, pp, va, vb, method = "Convolve")
}
